from concurrent.futures import ThreadPoolExecutor, as_completed
import time
import re
import requests

# Lazy imports to avoid circular import issues
# runCASSIA and runCASSIA_batch are imported inside functions that need them

def _get_runCASSIA():
    """Lazy import of runCASSIA to avoid circular imports."""
    try:
        from CASSIA.engine.tools_function import runCASSIA
        return runCASSIA
    except ImportError:
        try:
            from ...engine.tools_function import runCASSIA
            return runCASSIA
        except ImportError:
            from tools_function import runCASSIA
            return runCASSIA

def _get_runCASSIA_batch():
    """Lazy import of runCASSIA_batch to avoid circular imports."""
    try:
        from CASSIA.engine.tools_function import runCASSIA_batch
        return runCASSIA_batch
    except ImportError:
        try:
            from ...engine.tools_function import runCASSIA_batch
            return runCASSIA_batch
        except ImportError:
            from tools_function import runCASSIA_batch
            return runCASSIA_batch

def _get_call_llm():
    """Lazy import of call_llm to avoid circular imports."""
    try:
        from CASSIA.core.llm_utils import call_llm
        return call_llm
    except ImportError:
        try:
            from ...core.llm_utils import call_llm
            return call_llm
        except ImportError:
            from llm_utils import call_llm
            return call_llm

def _get_get_top_markers():
    """Lazy import of get_top_markers to avoid circular imports."""
    try:
        from CASSIA.core.marker_utils import get_top_markers
        return get_top_markers
    except ImportError:
        try:
            from ...core.marker_utils import get_top_markers
            return get_top_markers
        except ImportError:
            from marker_utils import get_top_markers
            return get_top_markers



def runCASSIA_batch_n_times(n, marker, output_name="cell_type_analysis_results", model="google/gemini-2.5-flash-preview", temperature=0, tissue="lung", species="human", additional_info=None, celltype_column=None, gene_column_name=None, max_workers=10, batch_max_workers=5, provider="openrouter", max_retries=1, validator_involvement="v1"):
    """
    Run multiple batch cell type analyses in parallel.
    
    Args:
        n (int): Number of batch analyses to run
        marker: DataFrame or path to CSV file containing marker data
        output_name (str): Base name for output files
        model (str): Model name to use
        temperature (float): Temperature parameter for the model
        tissue (str): Tissue type
        species (str): Species type
        additional_info (str): Additional information for analysis
        celltype_column (str): Name of column containing cell types
        gene_column_name (str): Name of column containing gene lists
        max_workers (int): Maximum number of workers for each batch
        batch_max_workers (int): Maximum number of concurrent batch runs
        provider (str): AI provider to use ('openai', 'anthropic', 'openrouter', or a custom URL)
        max_retries (int): Maximum number of retries for failed analyses
        validator_involvement (str): Validator involvement level ('v0' for high involvement, 'v1' for moderate involvement)
    
    Returns:
        None: Results are saved to files
    """
    def single_batch_run(i):
        output_json_name = f"{output_name}_{i+1}.json"
        print(f"Starting batch run {i+1}/{n}")
        start_time = time.time()
        runCASSIA_batch = _get_runCASSIA_batch()
        result = runCASSIA_batch(
            marker=marker,
            output_name=output_json_name,
            model=model,
            temperature=temperature,
            tissue=tissue,
            species=species,
            additional_info=additional_info,
            celltype_column=celltype_column,
            gene_column_name=gene_column_name,
            max_workers=max_workers,
            provider=provider,
            max_retries=max_retries,
            validator_involvement=validator_involvement
        )
        end_time = time.time()
        print(f"Finished batch run {i+1}/{n} in {end_time - start_time:.2f} seconds")
        return i, result, output_json_name

    all_results = []
    start_time = time.time()

    with ThreadPoolExecutor(max_workers=batch_max_workers) as executor:
        future_to_index = {executor.submit(single_batch_run, i): i for i in range(n)}
        
        for future in as_completed(future_to_index):
            index = future_to_index[future]
            try:
                index, result, output_json_name = future.result()
                all_results.append((index, result, output_json_name))
                print(f"Batch run {index+1}/{n} completed and saved to {output_json_name}")
            except Exception as exc:
                print(f'Batch run {index+1} generated an exception: {exc}')

    end_time = time.time()
    print(f"All {n} batch runs completed in {end_time - start_time:.2f} seconds")

    return None

    #return all_results



def run_single_analysis(args):
    index, tissue, species, additional_info, temperature, marker_list, model, provider, validator_involvement, use_reference = args
    print(f"Starting analysis {index+1}")
    start_time = time.time()
    try:
        runCASSIA = _get_runCASSIA()
        result = runCASSIA(
            tissue=tissue,
            species=species,
            additional_info=additional_info,
            temperature=temperature,
            marker_list=marker_list,
            model=model,
            provider=provider,
            validator_involvement=validator_involvement,
            use_reference=use_reference
        )
        end_time = time.time()
        print(f"Finished analysis {index+1} in {end_time - start_time:.2f} seconds")
        return index, result
    except Exception as e:
        print(f"Error in analysis {index+1}: {str(e)}")
        return index, None





def runCASSIA_n_times(n, tissue, species, additional_info, temperature, marker_list, model, max_workers=10, provider="openrouter", validator_involvement="v1", use_reference=False):
    """
    Run multiple cell type analyses in parallel.

    Args:
        n (int): Number of analyses to run
        tissue (str): Tissue type
        species (str): Species type
        additional_info (str): Additional information for analysis
        temperature (float): Temperature parameter for the model
        marker_list (list): List of markers to analyze
        model (str): Model name to use
        max_workers (int): Maximum number of parallel workers
        provider (str): AI provider to use ('openai', 'anthropic', 'openrouter', or a custom URL)
        validator_involvement (str): Validator involvement level
        use_reference (bool): Whether to use reference-based annotation for complex cases

    Returns:
        dict: Dictionary of analysis results indexed by iteration number (each result is a 3-tuple)
    """
    print(f"Starting {n} parallel analyses")
    start_time = time.time()

    results = {}

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        # Submit all tasks with the provider parameter
        future_to_index = {
            executor.submit(
                run_single_analysis,
                (i, tissue, species, additional_info, temperature, marker_list, model, provider, validator_involvement, use_reference)
            ): i for i in range(n)
        }
        
        for future in as_completed(future_to_index):
            index = future_to_index[future]
            try:
                index, result = future.result()
                if result:
                    results[index] = result
            except Exception as exc:
                print(f'Analysis {index+1} generated an exception: {exc}')
    
    end_time = time.time()
    print(f"All analyses completed in {end_time - start_time:.2f} seconds")
    return results







def parse_results_to_dict(result_string):
    """
    Robust parsing function that tries multiple regex patterns to handle different model outputs.
    """
    
    # Define multiple patterns to try, in order of preference
    patterns = [
        # Original strict pattern: result1:(cell_type, subtype) or result1:[cell_type, subtype]
        r"result(\d+):[\(\[]([^\]\)]+)[\)\]]",
        
        # Case insensitive with optional spaces: Result 1: (cell_type, subtype)
        r"(?i)result\s*(\d+)\s*:\s*[\(\[]([^\]\)]+)[\)\]]",
        
        # Simple numbered format: 1. (cell_type, subtype) or 1: (cell_type, subtype)
        r"(\d+)[\.\:]\s*[\(\[]([^\]\)]+)[\)\]]",
        
        # Even more flexible: any number followed by parentheses/brackets
        r"(\d+)[^\(\[]*[\(\[]([^\]\)]+)[\)\]]",
        
        # Pattern for results without explicit numbering but with parentheses
        r"[\(\[]([^,\]\)]+),\s*([^\]\)]+)[\)\]]",
        
        # Very loose pattern: find any content in parentheses with comma
        r"[\(\[]([^,\]\)]{3,}),\s*([^\]\)]{2,})[\)\]]"
    ]
    
    parsed_results = {}
    
    # Try each pattern until we find matches
    for pattern_idx, pattern in enumerate(patterns):
        matches = re.findall(pattern, result_string, re.IGNORECASE)
        
        if matches:
            print(f"Successfully parsed using pattern {pattern_idx + 1}: {pattern}")
            break
    
    # If no numbered patterns worked, try the non-numbered ones differently
    if not matches and pattern_idx >= 4:  # If we reached the non-numbered patterns
        # For non-numbered patterns, create sequential numbering
        sequential_matches = re.findall(patterns[4], result_string, re.IGNORECASE)
        if sequential_matches:
            matches = [(str(i+1), cell_types) for i, cell_types in enumerate(sequential_matches)]
            print(f"Created sequential numbering from pattern matches")
        else:
            sequential_matches = re.findall(patterns[5], result_string, re.IGNORECASE)
            if sequential_matches:
                matches = [(str(i+1), cell_types) for i, cell_types in enumerate(sequential_matches)]
                print(f"Created sequential numbering from loose pattern matches")
    
    # Parse each result
    for match in matches:
        try:
            if len(match) == 2:
                result_num, cell_types = match
            else:
                # Handle case where match might have different structure
                result_num = str(len(parsed_results) + 1)
                cell_types = str(match)
            
            # Handle different cell type formats
            if isinstance(cell_types, tuple):
                # If cell_types is already a tuple (from non-numbered patterns)
                cell_type_list = [str(ct).strip().strip("'\"") for ct in cell_types]
            else:
                # Split cell types, handling potential commas within cell type names
                # Try different splitting approaches
                if ',' in cell_types:
                    # Split by comma and clean up
                    cell_type_list = [ct.strip().strip("'\"") for ct in cell_types.split(',')]
                else:
                    # If no comma, try to split by other delimiters
                    for delimiter in [';', '|', ' and ', ' & ']:
                        if delimiter in cell_types:
                            cell_type_list = [ct.strip().strip("'\"") for ct in cell_types.split(delimiter)]
                            break
                    else:
                        # If no delimiter found, treat as single cell type
                        cell_type_list = [cell_types.strip().strip("'\"")]
            
            # Clean up cell types - remove common prefixes/suffixes that models might add
            cleaned_cell_types = []
            for ct in cell_type_list:
                ct = ct.strip()
                # Remove common model artifacts
                ct = re.sub(r'^(cell type:?\s*|type:?\s*)', '', ct, flags=re.IGNORECASE)
                ct = re.sub(r'\s*(cell type|type)\s*$', '', ct, flags=re.IGNORECASE)
                if ct:  # Only add non-empty cell types
                    cleaned_cell_types.append(ct)
            
            # Ensure we have at least two cell types (main and sub)
            while len(cleaned_cell_types) < 2:
                cleaned_cell_types.append("N/A")
            
            parsed_results[f"result{result_num}"] = tuple(cleaned_cell_types[:2])
            
        except Exception as e:
            print(f"Error parsing match {match}: {str(e)}")
            continue
    
    # If still no results, try one last desperate attempt
    if not parsed_results:
        print("No patterns matched. Attempting fallback parsing...")
        # Look for any two consecutive words that might be cell types
        fallback_pattern = r'\b([A-Z][a-z]+(?:\s+[A-Z][a-z]+)*)\s*,?\s*([A-Z][a-z]+(?:\s+[A-Z][a-z]+)*)\b'
        fallback_matches = re.findall(fallback_pattern, result_string)
        
        for i, (general, specific) in enumerate(fallback_matches[:5]):  # Limit to first 5 matches
            if general.lower() != specific.lower():  # Avoid duplicates
                parsed_results[f"result{i+1}"] = (general, specific)
        
        if parsed_results:
            print(f"Fallback parsing found {len(parsed_results)} potential cell type pairs")
    
    if not parsed_results:
        print("Warning: No results found to parse from input:")
        print(f"Input string: {result_string[:200]}..." if len(result_string) > 200 else result_string)
    
    return parsed_results



def extract_celltypes_from_llm(llm_response, provider="openai", single_analysis=False):
    """
    Unified function to extract cell type information from LLM responses.
    Handles differences in response format between different providers.
    
    Args:
        llm_response: The text response from the LLM
        provider: The provider that generated the response ("openai", "anthropic", "openrouter", or a custom URL)
        single_analysis: Whether this is from a single analysis (different return format)
        
    Returns:
        If single_analysis=False:
            general_celltype, sub_celltype, mixed_celltypes, consensus_score
        If single_analysis=True:
            general_celltype, sub_celltype, mixed_celltypes
    """
    import json
    import re
    
    # Default return values
    default_not_found = "Not found"
    default_score = 0
    
    # Try multiple extraction methods for all providers
    json_match = None
    
    # First try to extract JSON from <json> tags (common for Claude and some custom APIs)
    if not json_match:
        json_match = re.search(r'<json>(.*?)</json>', llm_response, re.DOTALL)
        
    # Try json> format (sometimes Claude drops the opening <)
        if not json_match:
            json_match = re.search(r'json>(.*?)</json>', llm_response, re.DOTALL)
    
    # Try markdown code blocks (common for OpenAI and some custom APIs)
    if not json_match:
        json_match = re.search(r'```json\n(.*?)\n```', llm_response, re.DOTALL)
        
    # Try code blocks with different formatting
    if not json_match:
        json_match = re.search(r'```\s*json\s*(.*?)\s*```', llm_response, re.DOTALL)
    
    # If still no match, try to find JSON object directly
    if not json_match:
        json_match = re.search(r'\{[\s\S]*\}', llm_response)
    
    # Process the match if found
    if json_match:
        try:
            # Extract the JSON content based on the matched pattern
            if '<json>' in llm_response or 'json>' in llm_response:
                json_str = json_match.group(1)
            elif '```json' in llm_response or '```' in llm_response:
                json_str = json_match.group(1)
            else:
                json_str = json_match.group(0)
                
            # Parse the JSON
            data = json.loads(json_str)
            
            final_results = data.get("final_results", [])
            mixed_celltypes = data.get("possible_mixed_celltypes", [])
            consensus_score = data.get("consensus_score", default_score)
            
            general_celltype = final_results[0] if len(final_results) > 0 else default_not_found
            sub_celltype = final_results[1] if len(final_results) > 1 else default_not_found
            
            # For single analysis, if general_celltype indicates no consensus, use mixed_celltypes
            if single_analysis and general_celltype.lower().startswith("no consensus"):
                general_celltype = ", ".join(mixed_celltypes)
            
            if single_analysis:
                return general_celltype, sub_celltype, mixed_celltypes
            else:
                return general_celltype, sub_celltype, mixed_celltypes, consensus_score
        except json.JSONDecodeError:
            print("Error decoding JSON from LLM response")
            print(f"Attempted to parse: {json_str if 'json_str' in locals() else 'No JSON found'}")
    else:
        print("No JSON data found in the LLM response")
        print(f"Full response: {llm_response[:500]}...")  # Print first 500 chars to avoid overwhelming output
    
    if single_analysis:
        return default_not_found, default_not_found, []
    else:
        return default_not_found, default_not_found, [], default_score


def consensus_similarity_flexible(results, main_weight=0.7, sub_weight=0.3):
    """
    Calculate consensus similarity from runCASSIA_n_times results.

    Args:
        results: dict where values are 3-tuples (analysis_result, conversation_history, reference_info)
                 analysis_result is a dict with 'main_cell_type' and 'sub_cell_types' keys
        main_weight: weight for main cell type agreement
        sub_weight: weight for sub cell type agreement

    Returns:
        tuple: (similarity_score, consensus_general, consensus_sub)
    """
    # Check if results is empty
    if not results:
        print("Warning: No results to calculate consensus similarity")
        return 0.0, "Unknown", "Unknown"

    # Extract cell types from the 3-tuple structure
    # result[0] is the analysis_result dict, result[1] is conversation_history, result[2] is reference_info
    general_types = [result[0]['main_cell_type'] for result in results.values() if result[0] and 'main_cell_type' in result[0]]
    sub_types = [', '.join(result[0].get('sub_cell_types', [])) for result in results.values() if result[0]]

    # Check if general_types or sub_types is empty
    if not general_types or not sub_types:
        print("Warning: Empty general_types or sub_types in consensus calculation")
        return 0.0, "Unknown", "Unknown"

    # Use Counter to get most common types
    from collections import Counter
    general_counter = Counter(general_types)
    sub_counter = Counter(sub_types)

    # Check if Counter is empty
    if not general_counter or not sub_counter:
        print("Warning: Empty counter in consensus calculation")
        return 0.0, "Unknown", "Unknown"

    consensus_general = general_counter.most_common(1)[0][0]
    consensus_sub = sub_counter.most_common(1)[0][0]

    total_score = 0
    for result in results.values():
        if not result[0]:
            continue
        result_general = result[0].get('main_cell_type', '')
        result_sub = ', '.join(result[0].get('sub_cell_types', []))

        if result_general == consensus_general:
            total_score += main_weight
        elif result_general == consensus_sub:
            total_score += main_weight * sub_weight

        if result_sub == consensus_sub:
            total_score += sub_weight
        elif result_sub == consensus_general:
            total_score += sub_weight * main_weight

    similarity_score = total_score / (len(results) * (main_weight + sub_weight))

    return similarity_score, consensus_general, consensus_sub


def consensus_similarity_from_tuples(parsed_results, main_weight=0.7, sub_weight=0.3):
    """
    Calculate consensus similarity from parsed results with simple tuple format.

    Args:
        parsed_results: dict where values are tuples (main_type, sub_type)
                       e.g., {"result1": ("Plasma Cell", "IgA-producing"), ...}
        main_weight: weight for main cell type agreement
        sub_weight: weight for sub cell type agreement

    Returns:
        tuple: (similarity_score, consensus_general, consensus_sub)
    """
    from collections import Counter

    if not parsed_results:
        return 0.0, "Unknown", "Unknown"

    # Extract cell types from the tuple structure
    general_types = []
    sub_types = []

    for key, value in parsed_results.items():
        if isinstance(value, tuple) and len(value) >= 2:
            main_type = str(value[0]).strip() if value[0] else "Unknown"
            sub_type = str(value[1]).strip() if value[1] else "Unknown"
            general_types.append(main_type)
            sub_types.append(sub_type)

    if not general_types or not sub_types:
        return 0.0, "Unknown", "Unknown"

    # Get consensus (most common) types
    general_counter = Counter(general_types)
    sub_counter = Counter(sub_types)

    consensus_general = general_counter.most_common(1)[0][0]
    consensus_sub = sub_counter.most_common(1)[0][0]

    # Calculate similarity score based on agreement with consensus
    total_score = 0
    for key, value in parsed_results.items():
        if isinstance(value, tuple) and len(value) >= 2:
            result_general = str(value[0]).strip() if value[0] else ""
            result_sub = str(value[1]).strip() if value[1] else ""

            if result_general == consensus_general:
                total_score += main_weight

            if result_sub == consensus_sub:
                total_score += sub_weight

    similarity_score = total_score / (len(parsed_results) * (main_weight + sub_weight))

    return similarity_score, consensus_general, consensus_sub




def agent_unification(prompt, system_prompt='''You are a careful professional biologist, specializing in single-cell RNA-seq analysis. You will be given a series of results from a celltype annotator. 

Your task is to unify all the celltype names, so that same celltype have the same name. Follow these rules EXACTLY:

1. The first letter of each word should be CAPITAL and others should be lowercase
2. Remove plurals (e.g., "cells" becomes "cell")  
3. Unify similar terms: "stem", "progenitor", and "immature" mean the same thing
4. Keep the exact same format as the input - only change the cell type names
5. ALWAYS wrap your output in <results></results> tags

CRITICAL: You MUST preserve the exact input format. If input is "result1:(cell1, cell2)" then output MUST be "result1:(Cell1, Cell2)"

Examples:

Input format:      
result1:(immune cell, t cell),result2:(Immune cells,t cell),result3:(T cell, cd8+ t cell)
                  
Output format:
<results>result1:(Immune Cell, T Cell),result2:(Immune Cell, T Cell),result3:(T Cell, Cd8+ T Cell)</results>

Another example:
                      
Input format:      
result1:(Hematopoietic stem/progenitor cells (HSPCs), T cell progenitors),result2:(Hematopoietic Progenitor cells,t cell),result3:(Hematopoietic progenitor cells, T cell)
                  
Output format:
<results>result1:(Hematopoietic Progenitor Cell, T Cell Progenitor),result2:(Hematopoietic Progenitor Cell, T Cell),result3:(Hematopoietic Progenitor Cell, T Cell)</results>

REMEMBER: Always use <results></results> tags around your output and maintain the exact same structure as the input.
''', model=None, provider="openai", temperature=0, deplural_only=False):
    """
    Unified function to call any LLM provider for cell type name unification.
    
    Args:
        prompt: The unification prompt containing cell type names to unify
        system_prompt: Instructions for the LLM
        model: Model to use (defaults to provider's default if None)
        provider: LLM provider ("openai", "anthropic", "openrouter", or a custom URL)
        temperature: Temperature for generation (0-1)
        deplural_only: If True, only removes plurals without other unification
    
    Returns:
        The unified cell type names as a string
    """
    # If deplural_only is True, override the system prompt
    if deplural_only:
        system_prompt = "Remove the plural for celltype name, keep the original input format."
    
    # Select default model based on provider if not specified
    if model is None:
        if provider == "openai":
            model = "gpt-4o"
        elif provider == "anthropic":
            model = "claude-3-5-sonnet-20241022"
        elif provider == "openrouter":
            model = "anthropic/claude-3.5-sonnet"
        elif provider.startswith("http"):
            # For custom API endpoints, use a default model if none specified
            model = model or "deepseek-chat"
    
    # Call LLM using the unified function
    call_llm = _get_call_llm()
    result = call_llm(
        prompt=prompt,
        provider=provider,
        model=model,
        temperature=temperature,
        system_prompt=system_prompt,
        max_tokens=4096  # Set to maximum allowed for most models
    )

    # Process the result for all providers to extract content from <results> tags
    # This ensures consistent behavior regardless of provider
    if not deplural_only:
        import re
        results_match = re.search(r'<results>(.*?)</results>', result, re.DOTALL)
        if results_match:
            return results_match.group(1)
    
    return result







def agent_judgement(prompt, system_prompt='''You are a careful professional biologist, specializing in single-cell RNA-seq analysis. You will be given a series of results from a cell type annotator.
Your task is to determine the consensus cell type. The first entry of each result is the general cell type and the second entry is the subtype. You should provide the final general cell type and the subtype. Considering all results, if you think there is very strong evidence of mixed cell types, please also list them. Please give your step-by-step reasoning and the final answer. We also want to know how robust our reuslts are, please give a consensus score ranging from 0 to 100 to show how similar the results are from different runs. $10,000 will be rewarded for the correct answer.
Output in JSON format:
{
"final_results": [
"General cell type here",
"Sub cell type here"
],
"possible_mixed_celltypes": [
"Mixed cell type 1 here",
"Mixed cell type 2 here"
],
"consensus_score": 0-100
}

'''
    , model=None, provider="openai", temperature=0, single_analysis=False, use_json_tags=False):
    """
    Unified function to call any LLM provider for cell type judgment.
    
    Args:
        prompt: The prompt containing cell type info for judgment
        system_prompt: Instructions for the LLM
        model: Model to use (defaults to provider's default if None)
        provider: LLM provider ("openai", "anthropic", "openrouter", or a custom URL)
        temperature: Temperature for generation (0-1)
        single_analysis: Whether this is a single analysis judgment 
        use_json_tags: Whether to add <json></json> tags around the output
        
    Returns:
        The judgment result as a string
    """
    # Select default model based on provider if not specified
    if model is None:
        if provider == "openai":
            model = "gpt-4o"
        elif provider == "anthropic":
            model = "claude-3-5-sonnet-20241022"
        elif provider == "openrouter":
            model = "anthropic/claude-3.5-sonnet"
        elif provider.startswith("http"):
            # For custom API endpoints, use a default model if none specified
            model = model or "deepseek-chat"
            
    # Modify system prompt for all providers to include JSON tags
    # This helps standardize the output format across all providers
    if use_json_tags or True:  # Always use JSON tags for consistency
        # Replace the JSON format instruction to include tags
        if "Output in JSON format:" in system_prompt:
            system_prompt = system_prompt.replace(
                "Output in JSON format:", 
                "Output in JSON format with tags:"
            ).replace(
                "consensus_score\": 0-100\n}", 
                "consensus_score\": 0-100\n}</json>"
            ).replace(
                "{\n", 
                "<json>{\n"
            )
    
    # Call LLM using the unified function
    call_llm = _get_call_llm()
    return call_llm(
        prompt=prompt,
        provider=provider,
        model=model,
        temperature=temperature,
        system_prompt=system_prompt
    )





def agent_unification_deplural(prompt, model="gpt-4o", provider="openai", temperature=0):
    """
    Wrapper around agent_unification that only removes plurals from cell type names.
    
    Args:
        prompt: The prompt containing cell type names to depluralize
        model: Model to use
        provider: LLM provider ("openai", "anthropic", "openrouter", or a custom URL)
        temperature: Temperature for generation (0-1)
        
    Returns:
        The depluralized cell type names as a string
    """
    return agent_unification(
        prompt=prompt,
        model=model,
        provider=provider,
        temperature=temperature,
        deplural_only=True
    )



def get_cell_type_info(cell_type_name, ontology="CL"):
    # Check if the cell type name contains "mixed" (case-insensitive)
    if "mixed" in cell_type_name.lower():
        return "mixed cell population", "mixed cell population"

    base_url = "https://www.ebi.ac.uk/ols/api/search"
    params = {
        "q": cell_type_name,
        "ontology": ontology,
        "rows": 1
    }
    
    try:
        response = requests.get(base_url, params=params)
        response.raise_for_status()
        data = response.json()
        
        if 'response' in data and 'docs' in data['response'] and data['response']['docs']:
            first_doc = data['response']['docs'][0]
            obo_id = first_doc.get('obo_id')
            label = first_doc.get('label')
            return obo_id, label
        else:
            return None, None
    
    except requests.RequestException:
        return None, None




def standardize_cell_types(input_string):
    """
    Robust standardization function that tries multiple parsing approaches.
    If parsing fails, it returns the original input_string.
    """
    # Remove all hyphens from the input string
    current_input_string = input_string.replace("-", " ")
    
    results = []
    parsing_successful = False

    # Try multiple regex patterns to extract results
    patterns = [
        # Original strict pattern with single quotes
        r"result\d+:\('([^']+)', '([^']+)'\)",
        # Pattern with double quotes
        r"result\d+:\(\"([^\"]+)\", \"([^\"]+)\"\)",
        # Pattern without quotes
        r"result\d+:\(([^,]+), ([^)]+)\)",
        # Case insensitive result pattern
        r"(?i)result\s*\d+\s*:\s*[\(\[]([^,\]\)]+),\s*([^\]\)]+)[\)\]]",
        # Very flexible pattern
        r"(?i)\d+[^\(\[]*[\(\[]([^,\]\)]+),\s*([^\]\)]+)[\)\]]"
    ]
    
    # Try each pattern until we find matches
    for pattern_idx, pattern in enumerate(patterns):
        results = re.findall(pattern, current_input_string)
        if results:
            print(f"Standardization: Successfully parsed using pattern {pattern_idx + 1}")
            parsing_successful = True
            break
    
    # If no pattern worked, try to use the robust parser
    if not parsing_successful:
        print(f"Standardization: No direct patterns matched, trying robust parser...")
        parsed_dict = parse_results_to_dict(current_input_string)
        
        if parsed_dict:
            # Convert parsed dictionary back to tuples
            results = [(cell_types[0], cell_types[1]) for cell_types in parsed_dict.values()]
            print(f"Standardization: Extracted {len(results)} results using robust parser")
            parsing_successful = True

    # If parsing failed after all attempts
    if not parsing_successful: # or not results
        print("Warning: No results found to standardize.")
        print(f"Input string for standardization was: {input_string[:200]}..." if len(input_string) > 200 else input_string)
        return input_string # Return the original input_string if parsing failed
    
    standardized_results_list = []
    for i, (general_type, specific_type) in enumerate(results, 1):
        # Clean up the cell types before ontology lookup
        general_type = general_type.strip().strip("'\"")
        specific_type = specific_type.strip().strip("'\"")
        
        # Search for standardized names
        _, general_label = get_cell_type_info(general_type)
        _, specific_label = get_cell_type_info(specific_type)
        
        # Use original names if no standardized names found
        general_label = general_label or general_type
        specific_label = specific_label or specific_type
        
        standardized_results_list.append(f"result{i}:('{general_label}', '{specific_label}')")
    
    return ",".join(standardized_results_list)


import pandas as pd
import glob
from collections import defaultdict


def organize_batch_results(marker, file_pattern, celltype_column=None):
    # Read marker data
    if isinstance(marker, pd.DataFrame):
        df = marker.copy()
    elif isinstance(marker, str):
        df = pd.read_csv(marker)
    else:
        raise ValueError("marker must be either a pandas DataFrame or a string path to a CSV file")

    # Only process with get_top_markers if more than 2 columns
    if len(df.columns) > 2:
        get_top_markers = _get_get_top_markers()
        marker = get_top_markers(df, n_genes=50)
    else:
        marker = df  # Use the DataFrame directly if it has 2 or fewer columns
        
    # If celltype_column is not provided, use the first column
    if celltype_column is None:
        celltype_column = marker.columns[0]
    
    marker_celltype = marker[celltype_column]

    # Use glob to find all matching files
    file_list = sorted(glob.glob(file_pattern))

    # Initialize a defaultdict to store results for each cell type
    results = defaultdict(list)

    # Loop through each file (round of results)
    for file in file_list:
        df = pd.read_csv(file)
        
        # Loop through each cell type
        # Determine the cluster column name (new name first, then fall back to old name)
        cluster_col = 'Cluster ID' if 'Cluster ID' in df.columns else 'True Cell Type'
        for celltype in marker_celltype:
            # Find the row for the current cell type
            row = df[df[cluster_col] == celltype]
            
            if not row.empty:
                # Extract the predicted general cell type (second column)
                predicted_general = row.iloc[0, 1]
                
                # Extract the first predicted subtype (first element in the third column)
                predicted_subtypes = row.iloc[0, 2]
                first_subtype = predicted_subtypes.split(',')[0].strip() if pd.notna(predicted_subtypes) else 'N/A'
                
                # Append the results as a tuple
                results[celltype].append((predicted_general, first_subtype))

    # Convert the defaultdict to a regular dict for easier handling
    organized_results = dict(results)
    
    return organized_results


def process_cell_type_variance_analysis_batch(results, model=None, provider="openai", temperature=0.0, main_weight=0.5, sub_weight=0.5, include_ontology=True):
    """
    Unified function to process cell type variance analysis for batch results.
    
    Args:
        results: The formatted batch results string
        model: Model to use (defaults to provider's default if None)
        provider: LLM provider ("openai", "anthropic", "openrouter", or a custom URL)
        temperature: Temperature for generation (0-1)
        main_weight: Weight for the main cell type in similarity calculation
        sub_weight: Weight for the sub cell type in similarity calculation
        include_ontology: Whether to include ontology-based standardization (additional analysis)
        
    Returns:
        Dictionary containing processed results
    """
    
    # Default model based on provider if not specified
    if model is None:
        if provider == "openai":
            model = "gpt-4o"
        elif provider == "anthropic":
            model = "claude-3-5-sonnet-20241022"
        elif provider == "openrouter":
            model = "anthropic/claude-3.5-sonnet"
        elif provider.startswith("http"):
            # For custom API endpoints, use a default model if none specified
            model = model or "deepseek-chat"
    
    # Extract and format results using the unified agent_unification function
    try:
        results_unification_llm = agent_unification(
            prompt=results,
            model=model,
            provider=provider,
            temperature=temperature
        )
    except Exception as e:
        print(f"Error in agent_unification: {str(e)}")
        results_unification_llm = results  # Fall back to original results
    
    # Process with ontology if requested
    ontology_results_available = False
    if include_ontology:
        try:
            # Depluralize using unified function
            results_depluar = agent_unification_deplural(
                prompt=results, # This 'results' is the original full batch string
                model=model,
                provider=provider,
                temperature=temperature
            )
            
            # Attempt to standardize cell types
            result_unified_oncology = standardize_cell_types(results_depluar)
            results_depluar_retry = None # Initialize results_depluar_retry to None
            
            # Check if standardization failed (returned input string unchanged)
            # and if the input string was non-trivial
            if result_unified_oncology == results_depluar and results_depluar and results_depluar.strip():
                print("Standardization failed on first attempt. Attempting re-generation of depluralized string and retrying standardization...")
                # Re-run depluralization
                results_depluar_retry = agent_unification_deplural(
                    prompt=results, # Use the same original prompt
                    model=model,
                    provider=provider,
                    temperature=temperature
                )
                # Retry standardization with the new depluralized string
                result_unified_oncology = standardize_cell_types(results_depluar_retry)
                # If it still fails, result_unified_oncology will be results_depluar_retry. 
                # The subsequent logic will handle it as a failed standardization.

            # Check if any actual standardization occurred or if it's still the raw depluralized string
            if result_unified_oncology != results_depluar and \
               (results_depluar_retry is None or result_unified_oncology != results_depluar_retry):
                 ontology_results_available = True
            else:
                 # This means even after a potential retry, standardization didn't change the string,
                 # or the initial standardization worked but produced the same string (e.g., already standardized)
                 # or it failed and returned the input string.
                 # We should consider it not truly "available" if it's identical to the unstandardized depluralized form
                 # unless results_depluar itself was already in the final standardized format.
                 # A simple check is if it contains "result1:(" which is our target format.
                 if "result1:(" in result_unified_oncology:
                     ontology_results_available = True
                 else:
                     ontology_results_available = False
                     print("Ontology standardization did not produce a parseable result or made no changes.")

        except Exception as e:
            print(f"Error in ontology processing: {str(e)}")
            result_unified_oncology = results  # Fall back to original results
    
    # Consensus judgment using unified function - always use JSON tags for consistency
    try:
        result_consensus_from_llm = agent_judgement(
            prompt=results_unification_llm,
            model=model,
            provider=provider,
            temperature=temperature,
            use_json_tags=True  # Always use JSON tags for all providers
        )
    except Exception as e:
        print(f"Error in agent_judgement: {str(e)}")
        result_consensus_from_llm = "{}"  # Empty JSON as fallback
    
    # Extract consensus celltypes using unified function
    try:
        general_celltype, sub_celltype, mixed_types, llm_generated_consensus_score_llm = extract_celltypes_from_llm(
            llm_response=result_consensus_from_llm,
            provider=provider
        )
    except Exception as e:
        print(f"Error extracting celltypes: {str(e)}")
        general_celltype, sub_celltype, mixed_types, llm_generated_consensus_score_llm = "Unknown", "Unknown", [], 0
    
    # Calculate similarity score
    try:
        parsed_results_llm = parse_results_to_dict(results_unification_llm)
        consensus_score_llm, consensus_1_llm, consensus_2_llm = consensus_similarity_from_tuples(
            parsed_results_llm,
            main_weight=main_weight,
            sub_weight=sub_weight
        )
    except Exception as e:
        print(f"Error calculating similarity score: {str(e)}")
        consensus_score_llm, consensus_1_llm, consensus_2_llm = 0, "Unknown", "Unknown"
    
    # Build result dictionary
    result = {
        'general_celltype_llm': general_celltype,
        'sub_celltype_llm': sub_celltype,
        'mixed_celltypes_llm': mixed_types,
        'consensus_score_llm': consensus_score_llm,
        'llm_generated_consensus_score_llm': llm_generated_consensus_score_llm,
        'count_consensus_1_llm': consensus_1_llm,
        'count_consensus_2_llm': consensus_2_llm,
        'unified_results_llm': results_unification_llm,
        'result_consensus_from_llm': result_consensus_from_llm
    }
    
    # Add ontology-based results if included and available
    if include_ontology and ontology_results_available:
        try:
            # Get consensus from ontology-standardized results - always use JSON tags
            result_consensus_from_oncology = agent_judgement(
                prompt=result_unified_oncology,
                model=model,
                provider=provider,
                temperature=temperature,
                use_json_tags=True  # Always use JSON tags for all providers
            )
            
            # Extract ontology-based consensus
            general_celltype_oncology, sub_celltype_oncology, mixed_types_oncology, llm_generated_consensus_score_oncology = extract_celltypes_from_llm(
                llm_response=result_consensus_from_oncology,
                provider=provider
            )
            
            # Calculate ontology-based similarity score
            parsed_results_oncology = parse_results_to_dict(result_unified_oncology)
            consensus_score_oncology, consensus_1_oncology, consensus_2_oncology = consensus_similarity_from_tuples(
                parsed_results_oncology,
                main_weight=main_weight,
                sub_weight=sub_weight
            )
            
            # Add ontology-based results to the result dictionary
            result.update({
                'general_celltype_oncology': general_celltype_oncology,
                'sub_celltype_oncology': sub_celltype_oncology,
                'mixed_types_oncology': mixed_types_oncology,
                'consensus_score_oncology': consensus_score_oncology,
                'llm_generated_consensus_score_oncology': llm_generated_consensus_score_oncology,
                'count_consensus_1_oncology': consensus_1_oncology,
                'count_consensus_2_oncology': consensus_2_oncology,
                'unified_results_oncology': result_unified_oncology,
                'result_consensus_from_oncology': result_consensus_from_oncology
            })
        except Exception as e:
            print(f"Error in ontology-based consensus processing: {str(e)}")
            # Add default values for ontology results
            result.update({
                'general_celltype_oncology': "Error in processing",
                'sub_celltype_oncology': "Error in processing",
                'mixed_types_oncology': [],
                'consensus_score_oncology': 0,
                'llm_generated_consensus_score_oncology': 0,
                'count_consensus_1_oncology': "Unknown",
                'count_consensus_2_oncology': "Unknown",
                'unified_results_oncology': result_unified_oncology,
                'result_consensus_from_oncology': "{}"
            })
    
    return result


def process_cell_type_results(organized_results, max_workers=10, model="google/gemini-2.5-flash-preview", provider="openrouter", main_weight=0.5, sub_weight=0.5, temperature=0.0):
    processed_results = {}
    
    def process_single_celltype(celltype, predictions):
        # Filter out 'N/A' predictions
        valid_predictions = [pred for pred in predictions if pred != ('N/A', 'N/A')]
        
        if not valid_predictions:
            return celltype, {
                'error': 'No valid predictions',
                'input': predictions
            }

        formatted_predictions = [f"result{i+1}:{pred}" for i, pred in enumerate(valid_predictions)]
        formatted_string = ",".join(formatted_predictions)

        # Use only the unified function
        result = process_cell_type_variance_analysis_batch(
            formatted_string,
            model=model,
            provider=provider,
            main_weight=main_weight,
            sub_weight=sub_weight,
            temperature=temperature
        )
        
        return celltype, result
    
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_celltype = {executor.submit(process_single_celltype, celltype, predictions): celltype 
                              for celltype, predictions in organized_results.items()}
        
        for future in as_completed(future_to_celltype):
            celltype = future_to_celltype[future]
            celltype, result = future.result()
            processed_results[celltype] = result
    
    return processed_results


# Update the function call
def create_and_save_results_dataframe(processed_results, organized_results, output_name='processed_cell_type_results'):
    """
    Create a DataFrame from processed results and save it to a CSV file.
    
    Args:
    processed_results (dict): Dictionary of processed results by cell type.
    organized_results (dict): Dictionary of original results by cell type.
    output_name (str): Base name for the output file (without extension)
    
    Returns:
    pd.DataFrame: Processed results in a DataFrame.
    """
    # Add .csv extension if not present
    output_csv = output_name if output_name.lower().endswith('.csv') else f"{output_name}.csv"
    
    # Create a list to store the data for each row
    data = []
    
    # Check if processed_results is empty
    if not processed_results:
        # Create an empty DataFrame with minimal columns
        df = pd.DataFrame(columns=['Cell Type', 'Status'])
        df.to_csv(output_csv, index=False)
        print(f"Empty results saved to {output_csv}")
        return df
    
    for celltype, result in processed_results.items():
        row_data = {
            'Cell Type': celltype,
            'General Cell Type LLM': result.get('general_celltype_llm', 'Not available'),
            'Sub Cell Type LLM': result.get('sub_celltype_llm', 'Not available'),
            'Mixed Cell Types LLM': ', '.join(result.get('mixed_celltypes_llm', [])),
            'General Cell Type Oncology': result.get('general_celltype_oncology', 'Not available'),
            'Sub Cell Type Oncology': result.get('sub_celltype_oncology', 'Not available'),
            'Mixed Cell Types Oncology': ', '.join(result.get('mixed_types_oncology', [])),
            'Similarity Score LLM': result.get('consensus_score_llm', 'Not available'),
            'Similarity Score Oncology': result.get('consensus_score_oncology', 'Not available'),
            'LLM Generated Consensus Score LLM': result.get('llm_generated_consensus_score_llm', 'Not available'),
            'LLM Generated Consensus Score Oncology': result.get('llm_generated_consensus_score_oncology', 'Not available'),
            'Count Consensus General Type LLM': result.get('count_consensus_1_llm', 'Not available'),
            'Count Consensus Sub Type LLM': result.get('count_consensus_2_llm', 'Not available'),
            'Count Consensus General Type Oncology': result.get('count_consensus_1_oncology', 'Not available'),
            'Count Consensus Sub Type Oncology': result.get('count_consensus_2_oncology', 'Not available'),
            'Unified Results LLM': result.get('unified_results_llm', 'Not available'),
            'Unified Results Oncology': result.get('unified_results_oncology', 'Not available'),
            'Consensus Result LLM': result.get('result_consensus_from_llm', 'Not available'),
            'Consensus Result Oncology': result.get('result_consensus_from_oncology', 'Not available'),
            'Original Non-Unified Results': ','.join([f"result{i+1}:{pred}" for i, pred in enumerate(organized_results.get(celltype, []))])
        }
        
        # Add original results
        original_results = organized_results.get(celltype, [])
        for i, (gen, sub) in enumerate(original_results, 1):
            row_data[f'Original General Type {i}'] = gen
            row_data[f'Original Sub Type {i}'] = sub
        
        data.append(row_data)

    # Create the DataFrame
    df = pd.DataFrame(data)

    # Check if DataFrame is empty
    if df.empty:
        df.to_csv(output_csv, index=False)
        print(f"Empty results saved to {output_csv}")
        return df
        
    # Define expected columns
    fixed_columns = ['Cell Type', 
                     'General Cell Type LLM', 'Sub Cell Type LLM', 'Mixed Cell Types LLM',
                     'General Cell Type Oncology', 'Sub Cell Type Oncology', 'Mixed Cell Types Oncology',
                     'Similarity Score LLM', 'Similarity Score Oncology',
                     'LLM Generated Consensus Score LLM', 'LLM Generated Consensus Score Oncology',
                     'Count Consensus General Type LLM', 'Count Consensus Sub Type LLM',
                     'Count Consensus General Type Oncology', 'Count Consensus Sub Type Oncology',
                     'Unified Results LLM', 'Unified Results Oncology',
                     'Consensus Result LLM', 'Consensus Result Oncology',
                     'Original Non-Unified Results']
    
    # Filter fixed columns to only include those that exist in the DataFrame
    available_fixed_columns = [col for col in fixed_columns if col in df.columns]
    
    # Get original columns
    original_columns = [col for col in df.columns if col.startswith('Original') and col != 'Original Non-Unified Results']
    
    # Reorder columns if possible
    if available_fixed_columns:
        df = df[available_fixed_columns + original_columns]

    # Save the DataFrame to a CSV file
    df.to_csv(output_csv, index=False)
    print(f"Results saved to {output_csv}")

    return df


def runCASSIA_similarity_score_batch(marker, file_pattern, output_name, celltype_column=None, max_workers=10, model="google/gemini-2.5-flash-preview", provider="openrouter", main_weight=0.5, sub_weight=0.5, temperature=0.0, generate_report=True, report_output_path=None):
    """
    Process batch results and save them to a CSV file, measuring the time taken.

    Args:
    marker_file_path (str): Path to the marker file.
    file_pattern (str): Path to pattern of result files.
    output_csv_name (str): Name of the output CSV file.
    celltype_column (str): Name of the column containing cell types in the marker file.
    max_workers (int): Maximum number of workers for parallel processing.
    model (str): Model name to use for LLM calls.
    provider (str): LLM provider ("openai", "anthropic", "openrouter", or a custom URL).
    main_weight (float): Weight for the main cell type in similarity calculation.
    sub_weight (float): Weight for the sub cell type in similarity calculation.
    temperature (float): Temperature for the LLM calls.
    generate_report (bool): Whether to generate an HTML report (default: True).
    report_output_path (str): Path to save the HTML report (default: 'uq_batch_report.html').
    """

    # Organize batch results
    organized_results = organize_batch_results(
        marker=marker,
        file_pattern=file_pattern,
        celltype_column=celltype_column
    )

    # Process cell type results
    processed_results = process_cell_type_results(organized_results, max_workers=max_workers, model=model, provider=provider, main_weight=main_weight, sub_weight=sub_weight, temperature=temperature)

    # Create and save results dataframe
    create_and_save_results_dataframe(
        processed_results,
        organized_results,
        output_name=output_name
    )

    print(f"Similarity analysis completed: {output_name}")

    # Generate HTML report if requested
    if generate_report:
        try:
            from CASSIA.reports.generate_report_uncertainty import generate_uq_batch_html_report
            report_path = report_output_path or 'uq_batch_report.html'
            generate_uq_batch_html_report(
                processed_results=processed_results,
                organized_results=organized_results,
                output_path=report_path,
                model=model,
                provider=provider
            )
        except ImportError as e:
            print(f"Warning: Could not generate report - {e}")
        except Exception as e:
            print(f"Warning: Report generation failed - {e}")



def extract_cell_types_from_results_single(results):
    extracted_results = []
    for i in range(len(results)):
        if i in results and results[i] is not None:
            result = results[i][0]  # Accessing the first element of each result
            main_cell_type = result.get('main_cell_type', 'Unknown')
            sub_cell_types = result.get('sub_cell_types', [])
            first_sub_cell_type = sub_cell_types[0] if sub_cell_types else 'None'
            extracted_results.append((main_cell_type, first_sub_cell_type))
        else:
            extracted_results.append(('Failed', 'Failed'))
    return extracted_results


def parse_results_to_dict_single(results):
    return {f"result{i+1}": result for i, result in enumerate(results)}

def extract_celltypes_from_llm_single(llm_response, provider="openai"):
    """
    Wrapper around the unified extract_celltypes_from_llm function for single analysis.

    Args:
        llm_response: The text response from the LLM
        provider: The provider that generated the response ("openai", "anthropic", "openrouter", or a custom URL)

    Returns:
        general_celltype, sub_celltype, mixed_celltypes, llm_consensus_score
    """
    # Use single_analysis=False to get all 4 return values including consensus_score
    general_celltype, sub_celltype, mixed_celltypes, consensus_score = extract_celltypes_from_llm(
        llm_response=llm_response,
        provider=provider,
        single_analysis=False
    )
    return general_celltype, sub_celltype, mixed_celltypes, consensus_score

def consensus_similarity_flexible_single(results, main_weight=0.7, sub_weight=0.3):
    general_types = [result[0] for result in results.values()]
    sub_types = [result[1] for result in results.values()]
    
    consensus_general = max(set(general_types), key=general_types.count)
    consensus_sub = max(set(sub_types), key=sub_types.count)
    
    total_score = sum(
        (main_weight if result[0] == consensus_general else 0) +
        (sub_weight if result[1] == consensus_sub else 0)
        for result in results.values()
    )
    
    similarity_score = total_score / (len(results) * (main_weight + sub_weight))
    
    return similarity_score, consensus_general, consensus_sub

def agent_judgement_single(prompt, system_prompt, model="gpt-4o", provider="openai", temperature=0):
    """
    Wrapper around agent_judgement for single analysis operations.
    
    Args:
        prompt: The prompt containing cell type info for judgment
        system_prompt: Instructions for the LLM
        model: Model to use (defaults to gpt-4o for OpenAI)
        provider: LLM provider ("openai", "anthropic", "openrouter", or a custom URL)
        temperature: Temperature for generation (0-1)
        
    Returns:
        The judgment result as a string
    """
    return agent_judgement(
        prompt=prompt, 
        system_prompt=system_prompt,
        model=model,
        provider=provider,
        temperature=temperature,
        single_analysis=True,
        use_json_tags=True  # Always use JSON tags for all providers
    )

def get_cell_type_info_single(cell_type_name, ontology="CL"):
    base_url = "https://www.ebi.ac.uk/ols/api/search"
    params = {
        "q": cell_type_name,
        "ontology": ontology,
        "rows": 1
    }
    
    try:
        response = requests.get(base_url, params=params)
        response.raise_for_status()
        data = response.json()
        
        if 'response' in data and 'docs' in data['response'] and data['response']['docs']:
            first_doc = data['response']['docs'][0]
            obo_id = first_doc.get('obo_id')
            label = first_doc.get('label')
            return obo_id, label
        else:
            return None, None
    
    except requests.RequestException:
        return None, None

def standardize_cell_types_single(results):
    standardized_results = []
    for i, (general_type, specific_type) in enumerate(results, 1):
        # Search for standardized names
        _, general_label = get_cell_type_info_single(general_type)
        _, specific_label = get_cell_type_info_single(specific_type)
        
        # Use original names if no standardized names found
        general_label = general_label or general_type
        specific_label = specific_label or specific_type
        
        standardized_results.append(f"result{i}:('{general_label}', '{specific_label}')")
    
    return ",".join(standardized_results)


def runCASSIA_n_times_similarity_score(tissue, species, additional_info, temperature, marker_list, model="google/gemini-2.5-flash-preview", max_workers=10, n=3, provider="openrouter", main_weight=0.5, sub_weight=0.5, validator_involvement="v1", use_reference=False, generate_report=True, report_output_path=None):
    """
    Wrapper function for processing cell type analysis using any supported provider.

    Args:
        tissue (str): Tissue type
        species (str): Species type
        additional_info (str): Additional information for analysis
        temperature (float): Temperature parameter for the model
        marker_list (list): List of markers to analyze
        model (str): Model name to use
        max_workers (int): Maximum number of parallel workers
        n (int): Number of analysis iterations
        provider (str): AI provider to use ('openai', 'anthropic', 'openrouter', or a custom URL)
        main_weight (float): Weight for main cell type in similarity calculation
        sub_weight (float): Weight for sub cell type in similarity calculation
        validator_involvement (str): Validator involvement level
        use_reference (bool): Whether to use reference information
        generate_report (bool): Whether to generate an HTML report (default: False)
        report_output_path (str): Path to save the HTML report (default: 'uq_report.html')

    Returns:
        dict: Analysis results including consensus types, cell types, and scores
    """
    # System prompt for all providers
    system_prompt = '''You are a careful professional biologist, specializing in single-cell RNA-seq analysis. You will be given a series of results from a cell type annotator.
Your task is to determine the consensus cell type. The first entry of each result is the general cell type and the second entry is the subtype. You should provide the final general cell type and the subtype. Considering all results, if you think there is very strong evidence of mixed cell types, please also list them. Please give your step-by-step reasoning and the final answer. Also give a consensus score ranging from 0 to 100 to show how similar the results are. $10,000 will be rewarded for the correct answer.
Output in JSON format:
{
"final_results": [
"General cell type here",
"Sub cell type here"
],
"possible_mixed_celltypes": [
"Mixed cell type 1 here",
"Mixed cell type 2 here"
],
"consensus_score": 0-100
}

'''

    # Run initial analysis
    results = runCASSIA_n_times(n, tissue, species, additional_info, temperature, marker_list, model, max_workers=max_workers, provider=provider, validator_involvement=validator_involvement, use_reference=use_reference)
    results = extract_cell_types_from_results_single(results)
    
    # Standardize cell types
    standardized_results = standardize_cell_types_single(results)
    
    # Get consensus judgment using the unified function
    result_consensus = agent_judgement_single(
        prompt=standardized_results,
        system_prompt=system_prompt,
        model=model,
        provider=provider,
        temperature=0.3
    )
    
    # Extract consensus celltypes using the unified function (now returns 4 values including llm_consensus_score)
    general_celltype, sub_celltype, mixed_types, llm_consensus_score = extract_celltypes_from_llm_single(
        result_consensus, provider=provider
    )

    # Calculate similarity score
    parsed_results = parse_results_to_dict_single(results)
    consensus_score, consensus_1, consensus_2 = consensus_similarity_flexible_single(parsed_results,main_weight=main_weight,sub_weight=sub_weight)

    final_results = {
        'unified_results': standardized_results,
        'consensus_types': (consensus_1, consensus_2),
        'general_celltype_llm': general_celltype,
        'sub_celltype_llm': sub_celltype,
        'Possible_mixed_celltypes_llm': mixed_types,
        'llm_response': result_consensus,
        'consensus_score_llm': consensus_score,  # Similarity score (0-1)
        'llm_generated_consensus_score_llm': llm_consensus_score,  # LLM consensus score (0-100)
        'similarity_score': consensus_score,
        'original_results': results
    }

    # Generate HTML report if requested
    if generate_report:
        try:
            from CASSIA.reports.generate_report_uncertainty import generate_uq_html_report
            report_path = report_output_path or 'uq_report.html'
            generate_uq_html_report(
                results=final_results,
                output_path=report_path,
                tissue=tissue,
                species=species,
                model=model,
                n_iterations=n,
                marker_list=marker_list
            )
        except ImportError as e:
            print(f"Warning: Could not generate report - {e}")
        except Exception as e:
            print(f"Warning: Report generation failed - {e}")

    return final_results




import pandas as pd
import numpy as np
import json
import os
import requests
import re  # Import regex
from typing import List, Dict, Union, Tuple, Optional, Any

class LLMEvaluator:
    """
    A class to evaluate cell type annotations using LLM-based scoring.
    Compares annotated cell types (from LLMs) against gold standard annotations.
    """
    
    def __init__(self, api_key: str = None, model: str = "deepseek/deepseek-chat-v3-0324"):
        """
        Initialize the LLM evaluator.
        
        Args:
            api_key (str): OpenRouter API key. If None, will try to get from environment.
            model (str): Model to use for evaluation from OpenRouter.
        """
        self.api_key = api_key or os.environ.get("OPENROUTER_API_KEY")
        if not self.api_key:
            raise ValueError("OpenRouter API key must be provided or set as OPENROUTER_API_KEY environment variable")
        
        self.model = model
    
    def get_single_celltype_prompts(self, 
                                   predicted_celltype: str, 
                                   gold_standard: str, 
                                   tissue: str, 
                                   species: str) -> Tuple[str, str]:
        """
        Generate system and user prompts for evaluating a single cell type prediction.
        
        Args:
            predicted_celltype (str): The cell type predicted by the annotation system
            gold_standard (str): The gold standard annotation
            tissue (str): The tissue context
            species (str): The species context
            
        Returns:
            Tuple[str, str]: System prompt and user prompt
        """
        
        system_prompt = """You are an expert cell biologist tasked with evaluating the accuracy of cell type annotations. You will be given a set of predicted cell types and their corresponding gold standard (correct) annotations.
Evaluate how accurate and specific each prediction is compared to its gold standard, considering cell ontology relationships.

For each cell type, there is a natural hierarchy: every specific (sub)cell type belongs to a broader general cell type category. The general cell type represents the broader classification (e.g., "epithelial cell"), while the subcell type refers to the more specific identity. for example, if the predicted cell type is "club cell", its general cell type is "epithelial cell. Its subcell type is "club cell".

You will be tip 10000$ if you did a good job, or 100 grandma are going to die.

For each pair, score the prediction on a scale of 0-5 where:
- 5: 100% exact match to the gold standard, or if the predicted cell type is more specific than the gold standard, if the predicted cell type has more details than the gold standard. We consider it as extra detail provided, so it is still a 100% match.
example: if the gold standard is "epithelial cell" and the prediction is "basal cell", assign a score of 5.
- 4: General cell type is correct AND subtype is mostly correct but missing subtle details (e.g., activation state, exhaustion status, or minor variations like vascular endothelial cell vs. lymphatic endothelial cell).It is either incomplete or not correct subtype information.
- 3: The general cell type is correct, but the predicted subtype matches partially to the gold standard. However, the prediction is still closely related in the cell ontology.
For example, the prediction is a differentiated cell type while the gold standard is an undifferentiated or progenitor cell type (e.g., "muscle cell progenitor cell" vs. "muscle cell").
Another example: the predicted and gold standard cell types are derived from the same ancestor but belong to different lineages (e.g., "dendritic cell" vs. "macrophage"—both derived from monocytes).
- 2: If the general cell type is correct, then it is at least a 2, it is not higher because of the subtype is very far from the gold standard, two or more than two layers away in the ontology tree. For example, predicted is cd8 t cell but gold standard is monocyte. or predicted is macrophage but gold standard is neutrophil.
- 1: Incorrect general cell type, means the even the broad cell type is not correct but prediction is somewhat related to the actual cell type in the broader classification
- 0: Completely irrelevant prediction that makes no sense in relation to the gold standard


Additional Note:

If the gold standard is a broad or feature-based label (such as "proliferating cell"), and the prediction is a cell type commonly known to exhibit that feature (for example, "keratinocyte" is known to proliferate), assign a score of 4. for example predicted celltype NKT cell and gold standard is CD8+ Cytotoxic T cell, assign a score of 4.They are both t cell and share the cytotoxic feature.


If the gold standard is a broad or feature-based label (such as "proliferating cell"), and the prediction is a cell type commonly known to exhibit that feature (for example, "keratinocyte" is known to proliferate), assign a score of 4. for example predicted celltype NKT cell and gold standard is CD8+ Cytotoxic T cell, assign a score of 4.They are both t cell and share the cytotoxic feature.
"""
        
        user_prompt = f"""
Evaluate the following cell type annotation prediction against the gold standard:

Context:
- Tissue: {tissue}
- Species: {species}

Predicted cell type: {predicted_celltype}
Gold standard annotation: {gold_standard}

Please provide your score (0-5) and explanation, followed by a JSON in the format:
{{
  "score": X,
  "explanation": "Your explanation"
}}
"""
        
        return system_prompt, user_prompt
    
    def get_multiple_celltypes_prompts(self, 
                                      predicted_celltypes: List[str], 
                                      gold_standards: List[str], 
                                      tissue: str, 
                                      species: str) -> Tuple[str, str]:
        """
        Generate system and user prompts for evaluating multiple cell type predictions.
        
        Args:
            predicted_celltypes (List[str]): List of predicted cell types
            gold_standards (List[str]): List of corresponding gold standards
            tissue (str): The tissue context
            species (str): The species context
            
        Returns:
            Tuple[str, str]: System prompt and user prompt
        """
        
        system_prompt = """You are an expert cell biologist tasked with evaluating the accuracy of cell type annotations. You will be given a set of predicted cell types and their corresponding gold standard (correct) annotations.
Evaluate how accurate and specific each prediction is compared to its gold standard, considering cell ontology relationships.

For each cell type, there is a natural hierarchy: every specific (sub)cell type belongs to a broader general cell type category. The general cell type represents the broader classification (e.g., "epithelial cell"), while the subcell type refers to the more specific identity. for example, if the predicted cell type is "club cell", its general cell type is "epithelial cell. Its subcell type is "club cell".

You will be tip 10000$ if you did a good job, or 100 grandma are going to die.

For each pair, score the prediction on a scale of 0-5 where:
- 5: 100% exact match to the gold standard, or if the predicted cell type is more specific than the gold standard, if the predicted cell type has more details than the gold standard. We consider it as extra detail provided, so it is still a 100% match.
example: if the gold standard is "epithelial cell" and the prediction is "basal cell", assign a score of 5.
- 4: General cell type is correct AND subtype is mostly correct but missing subtle details (e.g., activation state, exhaustion status, or minor variations like vascular endothelial cell vs. lymphatic endothelial cell).It is either incomplete or not correct subtype information.
- 3: The general cell type is correct, but the predicted subtype matches partially to the gold standard. However, the prediction is still closely related in the cell ontology.
For example, the prediction is a differentiated cell type while the gold standard is an undifferentiated or progenitor cell type (e.g., "muscle cell progenitor cell" vs. "muscle cell").
Another example: the predicted and gold standard cell types are derived from the same ancestor but belong to different lineages (e.g., "dendritic cell" vs. "macrophage"—both derived from monocytes).
- 2: Only the general cell type is correct but the subtype is very far in the, two or more than two layers away in the ontology tree. For example, predicted is cd8 t cell but gold standard is monocyte. or predicted is macrophage but gold standard is neutrophil.
- 1: Incorrect general cell type, means the even the broad cell type is not correct but prediction is somewhat related to the actual cell type in the broader classification
- 0: Completely irrelevant prediction that makes no sense in relation to the gold standard

Your response must include:
1. Individual scores for each prediction (0-5)
2. Brief explanations for your scoring decisions, specifically referencing cell ontology relationships
3. A JSON-formatted result with the format:
{
  "individual_scores": [score1, score2, ...],
  "explanations": ["explanation1", "explanation2", ...]
}

Additional Note:

If the gold standard is a broad or feature-based label (such as "proliferating cell"), and the prediction is a cell type commonly known to exhibit that feature (for example, "keratinocyte" is known to proliferate), assign a score of 4. for example predicted celltype NKT cell and gold standard is CD8+ Cytotoxic T cell, assign a score of 4.They are both t cell and share the cytotoxic feature.

"""
        
        # Construct the pairs for evaluation
        pairs = []
        for i, (pred, gold) in enumerate(zip(predicted_celltypes, gold_standards)):
            # Explicitly format the string for clarity
            pair_string = f"Pair {i+1}:\nPredicted: {pred}\nGold standard: {gold}\n"
            pairs.append(pair_string)
        
        pairs_text = "\n".join(pairs)
        
        user_prompt = f"""
Evaluate the following cell type annotation predictions against their gold standards:

Context:
- Tissue: {tissue}
- Species: {species}

{pairs_text}

Please provide your evaluation with individual scores (0-5) and explanations, followed by a JSON in the specified format.
"""
        
        return system_prompt, user_prompt
    
    def evaluate_single_celltype(self, 
                                predicted_celltype: str, 
                                gold_standard: str, 
                                tissue: str, 
                                species: str) -> Dict[str, Any]:
        """
        Evaluate a single cell type annotation against a gold standard.
        
        Args:
            predicted_celltype (str): The cell type predicted by the annotation system
            gold_standard (str): The gold standard annotation
            tissue (str): The tissue context
            species (str): The species context
            
        Returns:
            Dict containing score and explanation
        """
        
        system_prompt, user_prompt = self.get_single_celltype_prompts(
            predicted_celltype, gold_standard, tissue, species
        )
        response = self._call_llm(system_prompt, user_prompt)
        result = self._extract_score_json(response)
        return result
    
    def evaluate_multiple_celltypes(self, 
                                   predicted_celltypes: List[str], 
                                   gold_standards: List[str], 
                                   tissue: str, 
                                   species: str) -> Dict[str, Any]:
        """
        Evaluate multiple cell type annotations against their gold standards.
        
        Args:
            predicted_celltypes (List[str]): List of predicted cell types
            gold_standards (List[str]): List of corresponding gold standards
            tissue (str): The tissue context
            species (str): The species context
            
        Returns:
            Dict containing individual scores and explanations
        """
        
        if len(predicted_celltypes) != len(gold_standards):
            raise ValueError("Length of predicted_celltypes must match length of gold_standards")
        
        system_prompt, user_prompt = self.get_multiple_celltypes_prompts(
            predicted_celltypes, gold_standards, tissue, species
        )
        response = self._call_llm(system_prompt, user_prompt)
        result = self._extract_multiple_scores_json(response)
        return result
    
    def batch_evaluate_from_dataframe(self, 
                                     df: pd.DataFrame, 
                                     predicted_col: str, 
                                     gold_col: str,
                                     tissue_col: str = None,
                                     species_col: str = None,
                                     default_tissue: str = "unknown",
                                     default_species: str = "human") -> pd.DataFrame:
        """
        Evaluate a batch of predictions from a dataframe.
        If the prediction and gold columns contain lists (or comma-separated strings), use the multiple celltype prompt.
        Otherwise, use the single celltype prompt.
        
        Args:
            df (pd.DataFrame): Dataframe containing predictions and gold standards
            predicted_col (str): Column name for predictions
            gold_col (str): Column name for gold standards
            tissue_col (str, optional): Column name for tissue context
            species_col (str, optional): Column name for species context
            default_tissue (str): Default tissue if tissue_col is None
            default_species (str): Default species if species_col is None
            
        Returns:
            pd.DataFrame with original data plus evaluation results
        """
        result_df = df.copy()
        result_df['evaluation_score'] = None
        result_df['evaluation_explanation'] = None
        
        for idx, row in df.iterrows():
            tissue = row.get(tissue_col, default_tissue) if tissue_col else default_tissue
            species = row.get(species_col, default_species) if species_col else default_species
            pred_val = row[predicted_col]
            gold_val = row[gold_col]
            # Detect if both are lists or comma-separated strings
            is_multi = False
            pred_list = None
            gold_list = None
            if isinstance(pred_val, list) and isinstance(gold_val, list):
                is_multi = True
                pred_list = pred_val
                gold_list = gold_val
            elif isinstance(pred_val, str) and isinstance(gold_val, str):
                if "," in pred_val and "," in gold_val:
                    pred_list = [p.strip() for p in pred_val.split(",")]
                    gold_list = [g.strip() for g in gold_val.split(",")]
                    if len(pred_list) == len(gold_list):
                        is_multi = True
            # Use multiple or single evaluation
            if is_multi:
                eval_result = self.evaluate_multiple_celltypes(
                    predicted_celltypes=pred_list,
                    gold_standards=gold_list,
                    tissue=tissue,
                    species=species
                )
                # Store mean score and explanations
                scores = eval_result.get('individual_scores', [])
                explanations = eval_result.get('explanations', [])
                mean_score = sum(scores)/len(scores) if scores else 0
                result_df.at[idx, 'evaluation_score'] = mean_score
                result_df.at[idx, 'evaluation_explanation'] = '\n'.join(explanations)
            else:
                eval_result = self.evaluate_single_celltype(
                    predicted_celltype=pred_val,
                    gold_standard=gold_val,
                    tissue=tissue,
                    species=species
                )
                result_df.at[idx, 'evaluation_score'] = eval_result.get('score', 0)
                result_df.at[idx, 'evaluation_explanation'] = eval_result.get('explanation', '')
        return result_df
    
    def _call_llm(self, system_prompt: str, user_prompt: str) -> str:
        """
        Call the LLM using OpenRouter API.
        
        Args:
            system_prompt (str): System prompt for the LLM
            user_prompt (str): User prompt for the LLM
            
        Returns:
            str: LLM response text
        """
        try:
            response = requests.post(
                url="https://openrouter.ai/api/v1/chat/completions",
                headers={
                    "Authorization": f"Bearer {self.api_key}",
                    "HTTP-Referer": "https://elliotxie.github.io/CASSIA/",
                    "X-Title": "CASSIA",
                    "Content-Type": "application/json"
                },
                json={
                    "model": self.model,
                    "temperature": 0,  # Use zero temperature for evaluation
                    "max_tokens": 2000,
                    "messages": [
                        {"role": "system", "content": system_prompt},
                        {"role": "user", "content": user_prompt}
                    ]
                }
            )
            
            # Check if request was successful
            if response.status_code == 200:
                response_data = response.json()
                return response_data['choices'][0]['message']['content']
            else:
                print(f"Error: OpenRouter API returned status code {response.status_code}")
                print(f"Response: {response.text}")
                return ''
                
        except Exception as e:
            print(f"Error making OpenRouter API request: {str(e)}")
            return ''
    
    def _extract_score_json(self, response: str) -> Dict[str, Any]:
        """
        Extract the score and explanation from the LLM response.
        Tries to find and parse a JSON object; falls back to regex if needed.
        
        Args:
            response (str): LLM response text
            
        Returns:
            Dict with score and explanation
        """
        try:
            # First try to find JSON content
            import re
            json_match = re.search(r'\{[\s\S]*?\}', response)
            if json_match:
                json_str = json_match.group(0)
                result = json.loads(json_str)
                
                # Validate that we have the expected fields
                if 'score' in result:
                    return result
            
            # Fallback to regex for score if JSON parsing failed
            score_match = re.search(r'score:?\s*(\d+)', response, re.IGNORECASE)
            if score_match:
                score = int(score_match.group(1))
                
                # Extract explanation if possible
                explanation_match = re.search(r'explanation:?\s*(.+?)(?:\n|$)', response, re.IGNORECASE | re.DOTALL)
                explanation = explanation_match.group(1).strip() if explanation_match else "No explanation provided"
                
                return {
                    "score": score,
                    "explanation": explanation
                }
            
            # If all else fails, return a default response
            return {
                "score": 0,
                "explanation": "Failed to extract score and explanation from LLM response"
            }
            
        except Exception as e:
            print(f"Error extracting score from response: {str(e)}")
            return {
                "score": 0,
                "explanation": f"Error extracting score: {str(e)}"
            }
    
    def _extract_multiple_scores_json(self, response: str) -> Dict[str, Any]:
        """
        Extract multiple scores and explanations from the LLM response.
        
        Args:
            response (str): LLM response text
            
        Returns:
            Dict with individual scores and explanations
        """
        try:
            # First try to find JSON content
            import re
            json_match = re.search(r'\{[\s\S]*?\}', response)
            if json_match:
                json_str = json_match.group(0)
                result = json.loads(json_str)
                
                # Validate that we have the expected fields
                if 'individual_scores' in result:
                    return result
            
            # If JSON parsing failed, return a default response
            return {
                "individual_scores": [],
                "explanations": []
            }
            
        except Exception as e:
            print(f"Error extracting scores from response: {str(e)}")
            return {
                "individual_scores": [],
                "explanations": []
            }

    # --- NEW 0-100 Similarity Prompts ---
    def get_single_similarity_prompts(self, 
                                     predicted_celltype: str, 
                                     gold_standard: str, 
                                     tissue: str, 
                                     species: str) -> Tuple[str, str]:
        """
        Generate system and user prompts for evaluating similarity (0-100 scale).
        """
        system_prompt = """You are an expert cell biologist. Comprehensively evaluate the similarity between two cell types. Give a score from 0-100. Do a good job and you will be tipped 10000$!
        
        Note: if the predicted celltype is more specific than the gold standard, the score should be 90. For example, if the predicted celltype is "CD4+ T cell" and the gold standard is "T cell", the score should be 90. If the the two celltypes share a general celltype, then it should at least be 50. Only give 0 score when the two celltypes are compeletly differnt, means that they are not even from the same general celltype.

response format:
<reasoning>
Provide your detailed biological reasoning for the score here. Explain the relationship (or lack thereof) between the two cell types.
</reasoning>
<score>
Provide the numeric similarity score (0-100) here.
</score>
"""
        user_prompt = f"""
Context:
- Tissue: {tissue}
- Species: {species}

Predicted cell type: {predicted_celltype}
Gold standard annotation: {gold_standard}
"""
        return system_prompt, user_prompt

    def get_multiple_similarity_prompts(self, 
                                       predicted_celltypes: List[str], 
                                       gold_standards: List[str], 
                                       tissue: str, 
                                       species: str) -> Tuple[str, str]:
        """
        Generate system and user prompts for evaluating similarity for multiple pairs (0-100 scale).
        """
        system_prompt = """You are an expert cell biologist. For each pair of predicted and gold standard cell types provided below, evaluate their similarity.

Provide a similarity score from 0 (completely dissimilar) to 100 (identical or the prediction is a direct, valid subtype of the gold standard) for each pair.

Your response MUST strictly follow this format, repeating the tags for EACH pair:
<reasoning>
Reasoning for Pair 1.
</reasoning>
<score>
Score for Pair 1 (0-100).
</score>
<reasoning>
Reasoning for Pair 2.
</reasoning>
<score>
Score for Pair 2 (0-100).
</score>
... etc. for all pairs.
"""
        # Construct the pairs for evaluation
        pairs = []
        for i, (pred, gold) in enumerate(zip(predicted_celltypes, gold_standards)):
            # Explicitly format the string for clarity
            pair_string = f"Pair {i+1}:\nPredicted: {pred}\nGold standard: {gold}\n"
            pairs.append(pair_string)
        pairs_text = "\n".join(pairs)

        user_prompt = f"""
Evaluate the similarity for the following cell type annotation pairs:

Context:
- Tissue: {tissue}
- Species: {species}

{pairs_text}

For EACH pair, provide your reasoning and score using the specified <reasoning> and <score> tags, repeating the tag pair for each input pair.
"""
        return system_prompt, user_prompt

    # --- NEW 0-100 Similarity Evaluation Methods ---
    def evaluate_single_similarity(self, 
                                  predicted_celltype: str, 
                                  gold_standard: str, 
                                  tissue: str, 
                                  species: str) -> Dict[str, Any]:
        """
        Evaluate similarity for a single pair (0-100 scale).
        """
        system_prompt, user_prompt = self.get_single_similarity_prompts(
            predicted_celltype, gold_standard, tissue, species
        )
        print("\n--- SYSTEM PROMPT (Single Similarity) ---\n" + system_prompt + "\n--- END SYSTEM PROMPT ---\n")
        print("\n--- USER PROMPT (Single Similarity) ---\n" + user_prompt + "\n--- END USER PROMPT ---\n")
        response = self._call_llm(system_prompt, user_prompt)
        print(f"\n--- LLM RAW RESPONSE (Single Similarity) ---\n{response}\n--- END RAW RESPONSE ---\n") # Print raw response
        result = self._extract_similarity_single(response)
        return result

    def evaluate_multiple_similarity(self, 
                                    predicted_celltypes: List[str], 
                                    gold_standards: List[str], 
                                    tissue: str, 
                                    species: str) -> Dict[str, Any]:
        """
        Evaluate similarity for multiple pairs (0-100 scale).
        """
        if len(predicted_celltypes) != len(gold_standards):
            raise ValueError("Length of predicted_celltypes must match length of gold_standards")
        
        system_prompt, user_prompt = self.get_multiple_similarity_prompts(
            predicted_celltypes, gold_standards, tissue, species
        )
        print("\n--- SYSTEM PROMPT (Multiple Similarity) ---\n" + system_prompt + "\n--- END SYSTEM PROMPT ---\n")
        print("\n--- USER PROMPT (Multiple Similarity) ---\n" + user_prompt + "\n--- END USER PROMPT ---\n")
        response = self._call_llm(system_prompt, user_prompt)
        print(f"\n--- LLM RAW RESPONSE (Multiple Similarity) ---\n{response}\n--- END RAW RESPONSE ---\n") # Print raw response
        result = self._extract_similarity_multiple(response)
        return result

    # --- NEW Similarity Parsers ---
    def _extract_similarity_single(self, response: str) -> Dict[str, Any]:
        """
        Extract similarity score (0-100) and reasoning using tags.
        """
        try:
            print(f"[DEBUG] Extracting from response of length {len(response)}")
            
            # More robust pattern that allows for whitespace and is case-insensitive
            reasoning_match = re.search(r'<reasoning>\s*(.*?)\s*</reasoning>', response, re.DOTALL | re.IGNORECASE)
            score_match = re.search(r'<score>\s*(\d+)\s*</score>', response, re.IGNORECASE)
            
            if reasoning_match:
                reasoning = reasoning_match.group(1).strip()
                print(f"[DEBUG] Found reasoning: {reasoning[:50]}...")
            else:
                print("[DEBUG] No reasoning match found")
                reasoning = "No reasoning provided"
                
            score = None
            if score_match:
                score_str = score_match.group(1).strip()
                print(f"[DEBUG] Found score: {score_str}")
                try:
                    score = int(score_str)
                    if not (0 <= score <= 100):
                        print(f"Warning: Score {score} out of range 0-100.")
                        # Optionally clamp score
                        score = max(0, min(100, score)) 
                except ValueError:
                    print(f"[DEBUG] Warning: Could not convert score '{score_str}' to integer.")
                    score = None # Treat non-integer score as None
            else:
                print("[DEBUG] No score match found")
                
            # Try a more lenient pattern if the strict one failed
            if score is None:
                print("[DEBUG] Trying lenient score pattern")
                lenient_score_match = re.search(r'<score>.*?(\d+).*?</score>', response, re.DOTALL | re.IGNORECASE)
                if lenient_score_match:
                    score_str = lenient_score_match.group(1).strip()
                    print(f"[DEBUG] Found score with lenient pattern: {score_str}")
                    try:
                        score = int(score_str)
                    except ValueError:
                        print(f"[DEBUG] Warning: Still could not convert score '{score_str}' to integer.")

            result = {
                "similarity_score": score,
                "similarity_reasoning": reasoning
            }
            print(f"[DEBUG] Final extracted result: {result}")
            return result
        except Exception as e:
            print(f"[DEBUG] Error extracting similarity: {str(e)}")
            return {
                "similarity_score": None,
                "similarity_reasoning": f"Error extracting: {str(e)}"
            }

    def _extract_similarity_multiple(self, response: str) -> Dict[str, Any]:
        """
        Extract multiple similarity scores (0-100) and reasonings using tags.
        """
        try:
            print(f"[DEBUG] Extracting multiple from response of length {len(response)}")
            
            # More robust pattern that allows for whitespace and is case-insensitive
            reasonings = [m.strip() for m in re.findall(r'<reasoning>\s*(.*?)\s*</reasoning>', response, re.DOTALL | re.IGNORECASE)]
            score_strs = [m.strip() for m in re.findall(r'<score>\s*(\d+)\s*</score>', response, re.IGNORECASE)]
            
            print(f"[DEBUG] Found {len(reasonings)} reasoning(s) and {len(score_strs)} score(s)")
            
            scores = []
            if len(reasonings) != len(score_strs):
                 print(f"[DEBUG] Warning: Mismatch between number of reasonings ({len(reasonings)}) and scores ({len(score_strs)}).")
                 # Attempt to parse scores anyway, matching by index
                 
            for i, score_str in enumerate(score_strs):
                score = None
                try:
                    score = int(score_str)
                    if not (0 <= score <= 100):
                         print(f"[DEBUG] Warning: Score {score} at index {i} out of range 0-100.")
                         # Optionally clamp score
                         score = max(0, min(100, score))
                except ValueError:
                    print(f"[DEBUG] Warning: Could not convert score '{score_str}' at index {i} to integer.")
                scores.append(score)
                
            # If no scores found with strict pattern, try more lenient pattern
            if not scores:
                print("[DEBUG] Trying lenient score pattern for multiple scores")
                lenient_score_strs = [m.strip() for m in re.findall(r'<score>.*?(\d+).*?</score>', response, re.DOTALL | re.IGNORECASE)]
                print(f"[DEBUG] Found {len(lenient_score_strs)} scores with lenient pattern")
                
                for i, score_str in enumerate(lenient_score_strs):
                    score = None
                    try:
                        score = int(score_str)
                        if not (0 <= score <= 100):
                            score = max(0, min(100, score))
                    except ValueError:
                        print(f"[DEBUG] Warning: Could not convert lenient score '{score_str}' at index {i} to integer.")
                    scores.append(score)
                
            # Pad shorter list if mismatch occurred
            min_len = min(len(reasonings), len(scores))
            
            result = {
                "similarity_scores": scores[:min_len],  # Return scores up to the minimum length
                "similarity_reasonings": reasonings[:min_len]  # Return reasonings up to the minimum length
            }
            print(f"[DEBUG] Final extracted results: {len(result['similarity_scores'])} scores, first score: {result['similarity_scores'][0] if result['similarity_scores'] else None}")
            return result
        except Exception as e:
            print(f"[DEBUG] Error extracting multiple similarity scores/reasonings: {str(e)}")
            return {
                "similarity_scores": [],
                "similarity_reasonings": []
            }

# Example functions to generate simulated data for testing

def generate_simulated_data(n_samples: int = 10) -> pd.DataFrame:
    """
    Generate simulated data for testing the evaluator with an ontology-based scoring system.
    
    Args:
        n_samples (int): Number of samples to generate
        
    Returns:
        pd.DataFrame: Dataframe with simulated data
    """
    species_list = ["human", "mouse", "rat"]
    tissue_list = ["brain", "lung", "liver", "kidney", "heart", "spleen"]
    
    # Gold standard cell types with ontology hierarchy
    # Format: [(gold_standard, general_type)]
    cell_types = [
        ("CD8+ cytotoxic T cell", "T cell"),
        ("CD4+ helper T cell", "T cell"),
        ("CD4+ regulatory T cell", "T cell"),
        ("Memory B cell", "B cell"),
        ("Plasma cell", "B cell"),
        ("M1 macrophage", "Macrophage"),
        ("M2 macrophage", "Macrophage"),
        ("Alveolar macrophage", "Macrophage"),
        ("Conventional dendritic cell", "Dendritic cell"),
        ("Plasmacytoid dendritic cell", "Dendritic cell"),
        ("Neutrophil", "Granulocyte"),
        ("Eosinophil", "Granulocyte"),
        ("NK cell", "Lymphoid cell"),
        ("Vascular endothelial cell", "Endothelial cell"),
        ("Lymphatic endothelial cell", "Endothelial cell"),
        ("Type I pneumocyte", "Epithelial cell"),
        ("Type II pneumocyte", "Epithelial cell"),
        ("Ciliated epithelial cell", "Epithelial cell"),
        ("Fibroblast", "Stromal cell"),
        ("Myofibroblast", "Stromal cell")
    ]
    
    # Map cell types to their lineage for score=3 examples
    lineage_map = {
        "T cell": ["NK cell", "B cell"],  # Close immune cell types (same lymphoid lineage)
        "B cell": ["T cell", "NK cell"],
        "Macrophage": ["Dendritic cell", "Monocyte"],  # Close myeloid cells
        "Dendritic cell": ["Macrophage", "Monocyte"],
        "Granulocyte": ["Monocyte", "Macrophage"],
        "Lymphoid cell": ["T cell", "B cell"],
        "Endothelial cell": ["Vascular endothelial cell", "Lymphatic endothelial cell"],
        "Epithelial cell": ["Type I pneumocyte", "Type II pneumocyte", "Ciliated epithelial cell"],
        "Stromal cell": ["Fibroblast", "Myofibroblast"]
    }
    
    # Generate data
    data = []
    for i in range(n_samples):
        # Select random gold standard
        gold_standard, general_type = cell_types[i % len(cell_types)]
        
        # Generate simulated prediction with varying accuracy based on the 0-5 scale
        accuracy_level = np.random.choice([
            "exact",     # Score 5: Exact match
            "subtle",    # Score 4: Subtle differences
            "close",     # Score 3: Close on ontology tree
            "general",   # Score 2: Only general type correct
            "distant",   # Score 1: Incorrect but related
            "nonsense"   # Score 0: Completely irrelevant
        ], p=[0.2, 0.2, 0.2, 0.2, 0.1, 0.1])
        
        # Generate prediction based on accuracy level
        if accuracy_level == "exact":
            # Score 5: Exact match
            pred = gold_standard
            score = 5
            
        elif accuracy_level == "subtle":
            # Score 4: Subtle differences in activation state or minor variations
            if "T cell" in gold_standard:
                if "regulatory" in gold_standard.lower():
                    pred = gold_standard.replace("regulatory", "suppressor")
                elif "cytotoxic" in gold_standard.lower():
                    pred = gold_standard.replace("cytotoxic", "effector")
                elif "helper" in gold_standard.lower():
                    pred = gold_standard.replace("helper", "CD4+")
                else:
                    pred = gold_standard + " (resting)"
            elif "macrophage" in gold_standard.lower():
                if "M1" in gold_standard:
                    pred = "Inflammatory macrophage"
                elif "M2" in gold_standard:
                    pred = "Tissue-resident macrophage"
                else:
                    pred = gold_standard + " (activated)"
            elif "endothelial" in gold_standard.lower():
                if "vascular" in gold_standard.lower():
                    pred = "Blood vessel endothelial cell"
                elif "lymphatic" in gold_standard.lower():
                    pred = "Lymph vessel endothelial cell"
                else:
                    pred = gold_standard + " (fenestrated)"
            else:
                # Add a minor state variation
                states = ["activated", "mature", "resting"]
                pred = np.random.choice(states) + " " + gold_standard
            score = 4
            
        elif accuracy_level == "close":
            # Score 3: Close on ontology tree (same lineage, different subtype)
            if general_type in lineage_map:
                similar_types = lineage_map[general_type]
                if similar_types:
                    pred = np.random.choice(similar_types)
                else:
                    pred = general_type + " (subtype unclear)"
            else:
                # Fallback if no specific mapping
                pred = general_type + " (variant)"
            score = 3
            
        elif accuracy_level == "general":
            # Score 2: Only general type correct, subtype far off
            if "T cell" in gold_standard:
                pred = "Myeloid cell"  # Far from T cell but still immune
            elif "macrophage" in gold_standard.lower() or "dendritic" in gold_standard.lower():
                pred = "Lymphocyte"  # Far from myeloid but still immune
            elif "endothelial" in gold_standard.lower():
                pred = "Mesenchymal cell"  # Different but still structural
            elif "epithelial" in gold_standard.lower():
                pred = "Connective tissue cell"  # Different but still tissue
            else:
                # Generic distant type in same system
                pred = general_type + "-like cell"
            score = 2
            
        elif accuracy_level == "distant":
            # Score 1: Incorrect general type but somewhat related
            if "T cell" in gold_standard or "B cell" in gold_standard or "NK" in gold_standard:
                pred = "Epithelial cell"  # Completely different lineage
            elif "macrophage" in gold_standard.lower() or "dendritic" in gold_standard.lower():
                pred = "Fibroblast"  # Different system
            elif "endothelial" in gold_standard.lower() or "epithelial" in gold_standard.lower():
                pred = "Hematopoietic cell"  # Different system
            else:
                # Generic distantly related cell
                distant_types = ["Adipocyte", "Melanocyte", "Chondrocyte", "Erythrocyte"]
                pred = np.random.choice(distant_types)
            score = 1
            
        else:  # nonsense
            # Score 0: Completely irrelevant
            nonsense_terms = [
                "Unknown structure", 
                "Cellular debris",
                "Extracellular matrix component",
                "Non-cellular feature",
                "Artifact",
                "In vitro cell line",
                "Undifferentiated precursor"
            ]
            pred = np.random.choice(nonsense_terms)
            score = 0
        
        # Create the data entry
        data.append({
            "species": np.random.choice(species_list),
            "tissue": np.random.choice(tissue_list),
            "gold_standard": gold_standard,
            "predicted_celltype": pred,
            "true_accuracy": score
        })
    
    return pd.DataFrame(data)

def generate_multiple_celltype_samples(n_samples: int = 5, n_types_per_sample: int = 3) -> List[Dict[str, Any]]:
    """
    Generate samples with multiple cell types for testing with ontology-based scoring.
    
    Args:
        n_samples (int): Number of samples to generate
        n_types_per_sample (int): Number of cell types per sample
        
    Returns:
        List[Dict]: List of samples with predicted and gold standard cell types
    """
    species_list = ["human", "mouse"]
    tissue_list = ["brain", "lung", "liver", "kidney"]
    
    # Gold standard cell types with ontology hierarchy
    # Format: [(gold_standard, general_type, lineage)]
    cell_types = [
        # Lymphoid cells
        ("CD8+ cytotoxic T cell", "T cell", "Lymphoid"),
        ("CD4+ helper T cell", "T cell", "Lymphoid"),
        ("CD4+ regulatory T cell", "T cell", "Lymphoid"),
        ("Memory B cell", "B cell", "Lymphoid"),
        ("Plasma cell", "B cell", "Lymphoid"),
        ("CD56+ NK cell", "NK cell", "Lymphoid"),
        ("CD56bright NK cell", "NK cell", "Lymphoid"),
        
        # Myeloid cells
        ("M1 macrophage", "Macrophage", "Myeloid"),
        ("M2 macrophage", "Macrophage", "Myeloid"),
        ("Alveolar macrophage", "Macrophage", "Myeloid"),
        ("Conventional dendritic cell", "Dendritic cell", "Myeloid"),
        ("Plasmacytoid dendritic cell", "Dendritic cell", "Myeloid"),
        ("Neutrophil", "Granulocyte", "Myeloid"),
        ("Eosinophil", "Granulocyte", "Myeloid"),
        
        # Epithelial cells
        ("Type I pneumocyte", "Pneumocyte", "Epithelial"),
        ("Type II pneumocyte", "Pneumocyte", "Epithelial"),
        ("Ciliated epithelial cell", "Airway epithelial cell", "Epithelial"),
        ("Goblet cell", "Secretory epithelial cell", "Epithelial"),
        
        # Endothelial cells
        ("Vascular endothelial cell", "Endothelial cell", "Endothelial"),
        ("Lymphatic endothelial cell", "Endothelial cell", "Endothelial"),
        
        # Stromal cells
        ("Fibroblast", "Stromal cell", "Mesenchymal"),
        ("Myofibroblast", "Stromal cell", "Mesenchymal"),
        ("Pericyte", "Perivascular cell", "Mesenchymal")
    ]
    
    samples = []
    for i in range(n_samples):
        # Select tissue and species for this sample
        tissue = np.random.choice(tissue_list)
        species = np.random.choice(species_list)
        
        # Select n_types_per_sample random cell types
        selected_cells = np.random.choice(len(cell_types), n_types_per_sample, replace=False)
        
        gold_standards = []
        predicted_types = []
        true_scores = []
        
        for cell_idx in selected_cells:
            # Get the gold standard info
            gold_standard, general_type, lineage = cell_types[cell_idx]
            gold_standards.append(gold_standard)
            
            # Generate prediction with varying accuracy
            accuracy_level = np.random.choice([
                "exact",     # Score 5: Exact match
                "subtle",    # Score 4: Subtle differences
                "close",     # Score 3: Close on ontology tree
                "general",   # Score 2: Only general type correct
                "distant",   # Score 1: Incorrect but related
                "nonsense"   # Score 0: Completely irrelevant
            ], p=[0.2, 0.2, 0.2, 0.2, 0.1, 0.1])
            
            if accuracy_level == "exact":
                # Score 5: Perfect match
                pred = gold_standard
                score = 5
                
            elif accuracy_level == "subtle":
                # Score 4: Subtle differences in activation state or minor variations
                if "T cell" in gold_standard:
                    if "regulatory" in gold_standard.lower():
                        pred = gold_standard.replace("regulatory", "suppressor")
                    elif "cytotoxic" in gold_standard.lower():
                        pred = gold_standard.replace("cytotoxic", "effector")
                    elif "helper" in gold_standard.lower():
                        pred = gold_standard.replace("helper", "CD4+")
                    else:
                        pred = gold_standard + " (resting)"
                elif "macrophage" in gold_standard.lower():
                    if "M1" in gold_standard:
                        pred = "Inflammatory macrophage"
                    elif "M2" in gold_standard:
                        pred = "Tissue-resident macrophage"
                    elif "alveolar" in gold_standard.lower():
                        pred = "Lung macrophage"
                    else:
                        pred = gold_standard + " (activated)"
                elif "NK cell" in gold_standard:
                    if "CD56bright" in gold_standard:
                        pred = "Cytokine-producing NK cell"
                    else:
                        pred = "Natural killer cell"
                elif "pneumocyte" in gold_standard.lower():
                    if "Type I" in gold_standard:
                        pred = "Squamous alveolar epithelial cell"
                    elif "Type II" in gold_standard:
                        pred = "Surfactant-producing epithelial cell"
                elif "dendritic" in gold_standard.lower():
                    if "conventional" in gold_standard.lower():
                        pred = "cDC"
                    elif "plasmacytoid" in gold_standard.lower():
                        pred = "pDC"
                else:
                    # Add a minor state variation
                    states = ["activated", "mature", "resting"]
                    pred = np.random.choice(states) + " " + gold_standard
                score = 4
                
            elif accuracy_level == "close":
                # Score 3: Close on ontology tree (same lineage, different subtype)
                if lineage == "Lymphoid":
                    if "T cell" in gold_standard:
                        other_options = ["CD8+ T cell", "CD4+ T cell", "Memory T cell", "Naive T cell", "γδ T cell"]
                        pred = np.random.choice([o for o in other_options if o != gold_standard])
                    elif "B cell" in gold_standard:
                        other_options = ["Naive B cell", "Memory B cell", "Germinal center B cell"]
                        pred = np.random.choice([o for o in other_options if o != gold_standard])
                    elif "NK" in gold_standard:
                        pred = "Innate lymphoid cell"
                    else:
                        pred = "Lymphocyte"
                elif lineage == "Myeloid":
                    if "macrophage" in gold_standard.lower():
                        pred = "Dendritic cell"
                    elif "dendritic" in gold_standard.lower():
                        pred = "Macrophage"
                    elif "neutrophil" in gold_standard.lower() or "eosinophil" in gold_standard.lower():
                        pred = "Granulocyte"
                    else:
                        pred = "Myeloid cell"
                elif lineage == "Epithelial":
                    if "pneumocyte" in gold_standard.lower():
                        pred = "Alveolar epithelial cell"
                    else:
                        pred = "Epithelial cell"
                elif lineage == "Endothelial":
                    if "vascular" in gold_standard.lower():
                        pred = "Endothelial cell"
                    elif "lymphatic" in gold_standard.lower():
                        pred = "Endothelial cell"
                    else:
                        pred = "Endothelial cell"
                elif lineage == "Mesenchymal":
                    if "fibroblast" in gold_standard.lower():
                        pred = "Stromal cell"
                    else:
                        pred = "Mesenchymal cell"
                else:
                    # Generic close type
                    pred = general_type
                score = 3
                
            elif accuracy_level == "general":
                # Score 2: Only general type correct, subtype far off
                if lineage == "Lymphoid":
                    if "T cell" in gold_standard or "B cell" in gold_standard:
                        pred = "Myeloid cell"  # Wrong lineage but still immune
                    else:
                        pred = "Leukocyte"
                elif lineage == "Myeloid":
                    pred = "Lymphocyte"  # Wrong lineage but still immune
                elif lineage == "Epithelial":
                    pred = "Mesenchymal cell"  # Different lineage
                elif lineage == "Endothelial":
                    pred = "Epithelial cell"  # Different but still structural
                elif lineage == "Mesenchymal":
                    pred = "Endothelial cell"  # Different but still structural
                else:
                    # Generic distant type
                    pred = "Cell of " + lineage + " origin"
                score = 2
                
            elif accuracy_level == "distant":
                # Score 1: Incorrect general type but somewhat related
                if lineage in ["Lymphoid", "Myeloid"]:
                    distant_options = ["Epithelial cell", "Fibroblast", "Endothelial cell"]
                    pred = np.random.choice(distant_options)
                elif lineage in ["Epithelial", "Endothelial", "Mesenchymal"]:
                    distant_options = ["Hematopoietic cell", "Immune cell", "Blood cell"]
                    pred = np.random.choice(distant_options)
                else:
                    # Generic distantly related cell
                    distant_types = ["Adipocyte", "Melanocyte", "Chondrocyte", "Erythrocyte"]
                    pred = np.random.choice(distant_types)
                score = 1
                
            else:  # nonsense
                # Score 0: Completely irrelevant
                nonsense_terms = [
                    "Unknown structure", 
                    "Cellular debris",
                    "Extracellular matrix component",
                    "Non-cellular feature",
                    "Artifact",
                    "In vitro cell line",
                    "Undifferentiated precursor"
                ]
                pred = np.random.choice(nonsense_terms)
                score = 0
            
            predicted_types.append(pred)
            true_scores.append(score)
        
        samples.append({
            "tissue": tissue,
            "species": species,
            "gold_standards": gold_standards,
            "predicted_celltypes": predicted_types,
            "true_scores": true_scores
        })
    
    return samples


# Command-line interface
def main():
    import argparse
    
    parser = argparse.ArgumentParser(description='Evaluate cell type annotations using LLMs on a 0-5 scale')
    subparsers = parser.add_subparsers(dest='command', help='Command to run')
    
    # Common arguments
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('--api-key', type=str, help='OpenRouter API key')
    parent_parser.add_argument('--model', type=str, default='anthropic/claude-3.5-sonnet', 
                              help='Model to use for evaluation')
    
    # Single cell type evaluation
    single_parser = subparsers.add_parser('single', parents=[parent_parser], 
                                       help='Evaluate a single cell type annotation (0-5 scale)')
    single_parser.add_argument('--predicted', type=str, required=True, 
                            help='Predicted cell type')
    single_parser.add_argument('--gold', type=str, required=True, 
                             help='Gold standard annotation')
    single_parser.add_argument('--tissue', type=str, default='unknown', 
                             help='Tissue context')
    single_parser.add_argument('--species', type=str, default='human', 
                             help='Species context')
    
    # Multiple cell types evaluation
    multi_parser = subparsers.add_parser('multiple', parents=[parent_parser], 
                                      help='Evaluate multiple cell type annotations (0-5 scale)')
    multi_parser.add_argument('--predicted', type=str, required=True, nargs='+', 
                           help='Predicted cell types (space-separated)')
    multi_parser.add_argument('--gold', type=str, required=True, nargs='+', 
                           help='Gold standard annotations (space-separated)')
    multi_parser.add_argument('--tissue', type=str, default='unknown', 
                           help='Tissue context')
    multi_parser.add_argument('--species', type=str, default='human', 
                           help='Species context')
    
    # Batch evaluation from CSV
    batch_parser = subparsers.add_parser('batch', parents=[parent_parser], 
                                      help='Evaluate batch from CSV file (0-5 scale)')
    batch_parser.add_argument('--input', type=str, required=True, 
                           help='Input CSV file path')
    batch_parser.add_argument('--output', type=str, required=True, 
                           help='Output CSV file path')
    batch_parser.add_argument('--predicted-col', type=str, required=True, 
                           help='Column name for predictions')
    batch_parser.add_argument('--gold-col', type=str, required=True, 
                           help='Column name for gold standards')
    batch_parser.add_argument('--tissue-col', type=str, 
                           help='Column name for tissue context')
    batch_parser.add_argument('--species-col', type=str, 
                           help='Column name for species context')
    batch_parser.add_argument('--default-tissue', type=str, default='unknown', 
                           help='Default tissue if not in CSV')
    batch_parser.add_argument('--default-species', type=str, default='human', 
                           help='Default species if not in CSV')
    
    # Simulate data
    sim_parser = subparsers.add_parser('simulate', 
                                     help='Generate simulated data for testing (with 0-5 score scale)')
    sim_parser.add_argument('--output', type=str, required=True, 
                          help='Output CSV file path')
    sim_parser.add_argument('--samples', type=int, default=10, 
                          help='Number of samples to generate')
    
    args = parser.parse_args()
    
    if args.command == 'single':
        evaluator = LLMEvaluator(api_key=args.api_key, model=args.model)
        result = evaluator.evaluate_single_celltype(
            predicted_celltype=args.predicted,
            gold_standard=args.gold,
            tissue=args.tissue,
            species=args.species
        )
        print(json.dumps(result, indent=2))
        
    elif args.command == 'multiple':
        if len(args.predicted) != len(args.gold):
            print("Error: Number of predicted cell types must match number of gold standards")
            return
        
        evaluator = LLMEvaluator(api_key=args.api_key, model=args.model)
        result = evaluator.evaluate_multiple_celltypes(
            predicted_celltypes=args.predicted,
            gold_standards=args.gold,
            tissue=args.tissue,
            species=args.species
        )
        print(json.dumps(result, indent=2))
        
    elif args.command == 'batch':
        try:
            df = pd.read_csv(args.input)
            evaluator = LLMEvaluator(api_key=args.api_key, model=args.model)
            result_df = evaluator.batch_evaluate_from_dataframe(
                df=df,
                predicted_col=args.predicted_col,
                gold_col=args.gold_col,
                tissue_col=args.tissue_col,
                species_col=args.species_col,
                default_tissue=args.default_tissue,
                default_species=args.default_species
            )
            result_df.to_csv(args.output, index=False)
            print(f"Batch evaluation completed. Results saved to {args.output}")
            
        except Exception as e:
            print(f"Error processing batch: {str(e)}")
            
    elif args.command == 'simulate':
        simulated_data = generate_simulated_data(args.samples)
        simulated_data.to_csv(args.output, index=False)
        print(f"Generated {args.samples} simulated samples. Saved to {args.output}")
        
    else:
        parser.print_help()

if __name__ == "__main__":
    # If run as a script, use the command-line interface
    main()

def calculate_evaluation_metrics(eval_df: pd.DataFrame, score_col: str = 'evaluation_score', 
                         true_score_col: str = None) -> Dict[str, float]:
    """
    Calculate metrics from batch evaluation results.
    
    Args:
        eval_df (pd.DataFrame): DataFrame with evaluation results
        score_col (str): Column name for evaluation scores (0-5 or 0-100 scale)
        true_score_col (str, optional): Column name for true scores if available
        
    Returns:
        Dict[str, float]: Dictionary with evaluation metrics
    """
    max_score = eval_df[score_col].max()
    is_similarity_scale = max_score > 10
    metrics = {
        'mean_score': eval_df[score_col].mean(),
        'median_score': eval_df[score_col].median(),
        'min_score': eval_df[score_col].min(),
        'max_score': eval_df[score_col].max(),
        'std_score': eval_df[score_col].std(),
        'count': len(eval_df),
    }
    if not is_similarity_scale:
        metrics.update({
            'perfect_ratio': (eval_df[score_col] == 5).mean(),
            'very_good_ratio': (eval_df[score_col] == 4).mean(),
            'good_ratio': (eval_df[score_col] == 3).mean(),
            'partial_ratio': (eval_df[score_col] == 2).mean(),
            'poor_ratio': (eval_df[score_col] == 1).mean(),
            'nonsensical_ratio': (eval_df[score_col] == 0).mean(),
        })
    # If true scores are available, calculate correlation and error metrics
    if true_score_col and true_score_col in eval_df.columns:
        metrics.update({
            'correlation': eval_df[[score_col, true_score_col]].corr().iloc[0, 1],
            'mae': np.abs(eval_df[score_col] - eval_df[true_score_col]).mean(),
            'rmse': np.sqrt(((eval_df[score_col] - eval_df[true_score_col]) ** 2).mean()),
        })
    return metrics



# Uncertainty Quantification Agent

from .Uncertainty_quantification import (
    runCASSIA_n_times,
    runCASSIA_batch_n_times,
    runCASSIA_similarity_score_batch,
    runCASSIA_n_times_similarity_score
)

__all__ = [
    'runCASSIA_n_times',
    'runCASSIA_batch_n_times',
    'runCASSIA_similarity_score_batch',
    'runCASSIA_n_times_similarity_score',
]



# CASSIA Agents Module
# Specialized agent modules

# Import from submodules
from .annotation_boost.annotation_boost import (
    runCASSIA_annotationboost,
    runCASSIA_annotationboost_additional_task,
    iterative_marker_analysis
)
from .annotation_boost import runCASSIA_super_annotation_boost, runSuperAnnotationBoost

from .uncertainty.Uncertainty_quantification import (
    runCASSIA_n_times,
    runCASSIA_batch_n_times,
    runCASSIA_similarity_score_batch,
    runCASSIA_n_times_similarity_score
)

from .merging.merging_annotation import merge_annotations, merge_annotations_all

from .subclustering.subclustering import runCASSIA_subclusters, annotate_subclusters, runCASSIA_n_subcluster

# Alias for backward compatibility
runCASSIA_subclustering = runCASSIA_subclusters

# Reference agent
from .reference_agent import ReferenceAgent, get_reference_content, format_reference_for_prompt

__all__ = [
    # Annotation boost
    'runCASSIA_annotationboost',
    'runCASSIA_annotationboost_additional_task',
    'iterative_marker_analysis',
    'runCASSIA_super_annotation_boost',
    'runSuperAnnotationBoost',
    # Uncertainty
    'runCASSIA_n_times',
    'runCASSIA_batch_n_times',
    'runCASSIA_similarity_score_batch',
    'runCASSIA_n_times_similarity_score',
    # Merging
    'merge_annotations',
    'merge_annotations_all',
    # Subclustering
    'runCASSIA_subclusters',
    'runCASSIA_subclustering',  # Alias
    'annotate_subclusters',
    'runCASSIA_n_subcluster',
    # Reference agent
    'ReferenceAgent',
    'get_reference_content',
    'format_reference_for_prompt',
]



"""
Reference Agent Module for CASSIA

Provides intelligent reference document retrieval and context injection
for cell type annotation tasks. The agent analyzes marker genes to:
1. Determine preliminary cell type classification
2. Assess complexity/ambiguity using LLM
3. Decide whether reference retrieval is needed
4. Select and extract relevant reference content
"""

from .reference_agent import ReferenceAgent, get_reference_content
from .complexity_scorer import assess_complexity
from .reference_selector import select_references
from .section_extractor import extract_sections, parse_markdown

__all__ = [
    'ReferenceAgent',
    'get_reference_content',
    'assess_complexity',
    'select_references',
    'extract_sections',
    'parse_markdown',
]



import pandas as pd

result_df = pd.read_csv("C:/Users/ellio/OneDrive - UW-Madison/CASSIA+/o4mini/combined_scores_o4mini_evaluation.csv")

    #group based on the Tissue	Species and then calcualte the average score in the score column，save that stat to a new csv file

result_df_stat= result_df.groupby(["Tissue", "Species"])["score"].mean().reset_index()

# Save a title row to the csv file, using the model name
model_name = "O4Mini"
output_path = "C:/Users/ellio/OneDrive - UW-Madison/CASSIA+/o4mini/combined_scores_o4mini_evaluation_average.csv"

with open(output_path, 'w', encoding='utf-8') as f:
    f.write(f"Model: {model_name}\n")
    result_df_stat.to_csv(f, index=False)
    

def process_multiple_files(file_model_pairs, output_txt_path):
    """
    Processes a list of (CSV file path, model name) pairs, computes grouped averages,
    and writes all results to a TXT file. Each section shows the model name, file name,
    grouped average scores, and overall average score for that model.
    """
    with open(output_txt_path, 'w', encoding='utf-8') as out_f:
        for file_path, model_name in file_model_pairs:
            try:
                # Read the CSV file
                df = pd.read_csv(file_path)
                
                # Calculate the overall average for this model (from the full CSV)
                overall_model_avg = df["score"].mean()
                
                # Calculate the grouped averages
                result_df_stat = df.groupby(["Tissue", "Species"])["score"].mean().reset_index()
                
                # Write model and file info
                out_f.write(f"Model: {model_name}\n")
                out_f.write(f"File: {file_path}\n")
                
                # Write the overall model average
                out_f.write(f"Overall average score for {model_name}: {overall_model_avg:.4f}\n\n")
                
                # Write the grouped statistics
                out_f.write("Grouped statistics:\n")
                out_f.write(result_df_stat.to_csv(index=False))
                out_f.write("\n\n")
                
            except Exception as e:
                out_f.write(f"Model: {model_name}\nFile: {file_path}\nError: {e}\n\n")

# Example usage:
file_model_list = [
    ("C:/Users/ellio/OneDrive - UW-Madison/CASSIA+/o4mini/combined_scores_o4mini_evaluation.csv", "O4Mini"),
    ("C:/Users/ellio/OneDrive - UW-Madison/CASSIA+/Gemini2.5pro_2/combined_scores_gemini-2.5pro2_evaluation.csv", "Gemini-2.5pro"),
    ("C:/Users/ellio/OneDrive - UW-Madison/CASSIA+/gemini2.5flash/combined_scores_gemini2.5flash_evaluation.csv", "Gemini2.5flash"),
    ("C:/Users/ellio/OneDrive - UW-Madison/CASSIA+/GPT4.1/combined_scores_GPT4.1_evaluation.csv", "GPT4.1"),
    ("C:/Users/ellio/OneDrive - UW-Madison/CASSIA+/claude3.7/combined_scores_claude3.7_evaluation.csv", "claude3.7")
]
process_multiple_files(file_model_list, "C:/Users/ellio/OneDrive - UW-Madison/CASSIA+/combined_results.txt")








# CASSIA - Cell Annotation with Semantic Similarity for Intelligent Analysis
# Root module with backward-compatible exports from reorganized submodules

__version__ = "0.3.5"

# =============================================================================
# BACKWARD COMPATIBILITY LAYER
# =============================================================================
# This file re-exports all public functions from the new organized submodules
# to maintain backward compatibility with existing code that uses:
#   from CASSIA import runCASSIA_batch
#   from CASSIA import set_api_key
# etc.

# Import logging configuration (errors visible by default)
from .core.logging_config import get_logger, set_log_level, warn_user

# -----------------------------------------------------------------------------
# CORE ANALYSIS FUNCTIONS (most commonly used)
# -----------------------------------------------------------------------------
from .engine.tools_function import (
    runCASSIA,
    runCASSIA_batch,
    runCASSIA_with_reference,
    runCASSIA_batch_with_reference,
    set_api_key,
    set_openai_api_key,
    set_anthropic_api_key,
    set_openrouter_api_key,
)

# Main function code exports
from .engine.main_function_code import *

# -----------------------------------------------------------------------------
# MARKER UTILITIES
# -----------------------------------------------------------------------------
from .core.marker_utils import loadmarker, list_available_markers, split_markers, get_top_markers

# -----------------------------------------------------------------------------
# LLM UTILITIES
# -----------------------------------------------------------------------------
from .core.llm_utils import call_llm

# -----------------------------------------------------------------------------
# MODEL SETTINGS (with fuzzy alias support)
# -----------------------------------------------------------------------------
from .core.model_settings import (
    ModelSettings,
    resolve_model_name,
    get_recommended_model,
    get_available_aliases,
    print_available_models,
    print_available_aliases
)

# -----------------------------------------------------------------------------
# VALIDATION AND EXCEPTIONS
# -----------------------------------------------------------------------------
from .core.exceptions import (
    CASSIAValidationError,
    MarkerValidationError,
    TemperatureValidationError,
    ProviderValidationError,
    ModelValidationError,
    TissueSpeciesValidationError,
    BatchParameterValidationError
)

# Alias for backward compatibility
CASSIAError = CASSIAValidationError

from .core.validation import (
    validate_marker_list,
    validate_temperature,
    validate_tissue,
    validate_species,
    validate_provider,
    validate_model,
    validate_marker_dataframe,
    validate_runCASSIA_inputs,
    validate_runCASSIA_batch_inputs,
    validate_runCASSIA_with_reference_inputs
)

# -----------------------------------------------------------------------------
# SCORING AND EVALUATION
# -----------------------------------------------------------------------------
from .evaluation.scoring import (
    runCASSIA_score_batch,
    score_annotation_batch,
    prompt_creator_score,
    extract_score_and_reasoning,
    score_single_analysis,
    process_single_row
)

from .evaluation.cell_type_comparison import compareCelltypes

# -----------------------------------------------------------------------------
# MULTI-MODEL COMPARISON
# -----------------------------------------------------------------------------
try:
    from .comparison import symphonyCompare, SymphonyCompare
except ImportError:
    pass  # Module may not be available

# -----------------------------------------------------------------------------
# REPORT GENERATION
# -----------------------------------------------------------------------------
try:
    from .reports.generate_reports import (
        generate_analysis_html_report,
        generate_html_report,
        process_single_report,
        generate_score_index_page,
        runCASSIA_generate_score_report,
        generate_subclustering_report,
        calculate_evaluation_metrics
    )
except ImportError:
    pass  # Module may not be available

try:
    from .reports.generate_batch_report import generate_batch_html_report_from_data
except ImportError:
    pass

try:
    from .reports.generate_report_uncertainty import generate_uq_html_report
except ImportError:
    pass

# -----------------------------------------------------------------------------
# PIPELINE
# -----------------------------------------------------------------------------
try:
    from .pipeline.pipeline import runCASSIA_pipeline
except ImportError:
    pass  # Module may not be available

# -----------------------------------------------------------------------------
# ANNOTATION BOOST AGENT
# -----------------------------------------------------------------------------
try:
    from .agents.annotation_boost.annotation_boost import (
        runCASSIA_annotationboost,
        runCASSIA_annotationboost_additional_task,
        iterative_marker_analysis
    )
except ImportError:
    pass  # Module may not be available

# -----------------------------------------------------------------------------
# MERGING AGENT
# -----------------------------------------------------------------------------
try:
    from .agents.merging.merging_annotation import merge_annotations, merge_annotations_all
    from .agents.merging.merging_annotation import _create_annotation_prompt, _parse_llm_response
except ImportError:
    pass  # Module may not be available

# -----------------------------------------------------------------------------
# UNCERTAINTY QUANTIFICATION AGENT
# -----------------------------------------------------------------------------
try:
    from .agents.uncertainty.Uncertainty_quantification import (
        runCASSIA_n_times,
        runCASSIA_batch_n_times,
        runCASSIA_similarity_score_batch,
        runCASSIA_n_times_similarity_score
    )
except ImportError:
    pass  # Module may not be available

# -----------------------------------------------------------------------------
# SUBCLUSTERING AGENT
# -----------------------------------------------------------------------------
try:
    from .agents.subclustering import (
        runCASSIA_subclusters,
        runCASSIA_subclustering,  # Alias from __init__.py
        runCASSIA_n_subcluster,
        annotate_subclusters
    )
except ImportError:
    pass  # Module may not be available

# -----------------------------------------------------------------------------
# REFERENCE AGENT (intelligent reference retrieval)
# -----------------------------------------------------------------------------
try:
    from .agents.reference_agent import (
        ReferenceAgent,
        get_reference_content,
        format_reference_for_prompt,
        assess_complexity,
        select_references
    )
    from .agents.reference_agent.complexity_scorer import (
        assess_complexity_step1,
        select_references_step2
    )
except ImportError:
    pass  # Reference agent module may not be available

# -----------------------------------------------------------------------------
# HYPOTHESIS GENERATION
# -----------------------------------------------------------------------------
try:
    from .hypothesis import (
        generate_hypothesis,
        process_marker_file,
        run_multi_analysis,
        summarize_runs,
        # Aliases
        runCASSIA_hypothesis,
        run_hypothesis_analysis,
        summarize_hypothesis_runs
    )
except ImportError:
    pass  # Module may not be available

# -----------------------------------------------------------------------------
# IMAGE ANALYSIS
# -----------------------------------------------------------------------------
try:
    from .imaging.llm_image import analyze_image, call_llm_with_image
except ImportError:
    pass  # Module may not be available

# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS
# -----------------------------------------------------------------------------
from .core.utils import safe_get, natural_sort_key, clean_conversation_history, write_csv
from .core.progress_tracker import BatchProgressTracker

# End of __init__.py
